"""This module defines a Bayesian probabilistic model of EMA data.
The model is learned using observed EMA recordings from all participants,
stored in an ema_data.EmaDataSet instance.

This model version uses a mixture of Gaussian distributions
for the parameter vectors in each population of potential respondents,
represented by one group of participants.
A separate GMM is learned for each population.

Individual parameter distributions are approximated by sampling.
The population mixture model is common prior for all individuals in
the group recruited from the same population.

The model can NOT use Bayesian sequential learning.
It must be trained in a single run from a single batch of EMA data,
which is usually quite sufficient for practical applications.

However, the learn(...) method can be called several times,
in case learning stopped before reaching convergence.

*** Class Overview:

EmaModel: Defines posterior distributions of situation probabilities,
    and locations of latent variables for perceptual attribute(s),
    learned from EMA data in an ema_data.EmaDataSet instance.
    The EMA study may include data from different groups of participants,
    represented by an ema_group.EmaGroupModel instance for each population / participant group.

ema_group.EmaGroupModel: Defines a GMM for ONE population,
    and contains all individual response-probability models,
    implemented by an ema_respondent.EmaRespondentModel instance for each participant,
    in ONE group of test participants, assumed recruited from the SAME population.

ema_respondent.EmaRespondentModel: Distribution of individual parameter vector
    assumed to determine the observed EMA data for ONE participant.
    Each EMA record specifies
    (1) a nominal (possibly multidimensional) Situation category, and
    (2) ordinal Ratings for zero, one, or more subjective Attributes.
    The parameter distribution is represented by an array xi with many samples.

ema_base.EmaParamBase: common properties defining
    indexing into array of parameter vectors, and
    a prior GaussianRV instance for all GMM components in all groups.

*** Model Theory: The present model is theoretically similar to
the model of the PairedCompCalc package, partly described in
A. Leijon, M. Dahlquist, and K. Smeds (2019):
Bayesian analysis of paired-comparison sound quality ratings. JASA 146(5):3174–3183.

EXCEPT for the present use of a mixture model for the population(s).

Detailed math documentation is presented in
A. Leijon, P. von Gablenz, I. Holube, J. Taghia, and K. Smeds.
Bayesian analysis of ecological momentary assessment (EMA) data
collected in adults before and after hearing rehabilitation.
Frontiers in Digital Health, 5(1100705), 2023.

*** Version History:
* Version 1.0.0:
2023-05-17, EmaModel.initialize(...) using n_participants_per_comp instead of max_n_comp.

* Version 0.9.3:
2022-08-20, ensure restrict_attributes=False, in case tied response scales
2022-06-xx, adapted to use pandas.DataFrame storage in ema_data.

* Version 0.8.3:
2022-03-08, minor cleanup

* Version 0.8.1
2022-02-26, cleanup logger output, cleanup comments
2022-02-26, class def EmaGroupModel -> module ema_group, EmaRespondentModel -> ema_respondent
2022-02-26, complete separate GMM for each group, GMM components -> EmaGroupModel property comp

* Version 0.8
2022-02-12, Changed VI factorization for better approximation,
    with individual indicators conditional on parameter samples,
    defining variational q(zeta_n, xi_n) = q(zeta_n | xi_n) q(xi_n)

* Version 0.7.2
2022-01-31, Different random Generator for each EmaRespondentModel sampler, spawned from same seed

* Version 0.7.1
2022-01-30, NO module-global random Generator object here
2022-01-30, seed input to EmaModel.initialize, for reproducible results
2022-01-29, random.Generator argument to EmaRespondentModel.initialize
2022-01-11, minor update EmaRespondentModel.adapt_xi, restrict_xi

* Version 0.7:
2021-12-15, new method EmaRespondentModel.mean_attribute_grade
2021-12-20, new method EmaRespondentModel.nap_diff, calculate NAP distance

* Version 0.6:
2021-12-06, Allow user control switches: restrict_attribute OR restrict_threshold
2021-12-08, restrict_attribute: sensory-variable location average forced -> 0.
            restrict_threshold: response-threshold median forced -> 0.

* Version 0.5:
2021-11-03, first functional version
2021-11-04, removed EmaGroupModel.initialize_weights and EmaRespondentModel.initialize_weights
2021-11-05, Hamiltonian sampler -> EmaRespondentModel property -> better stepsize optimization
2021-11-07, store EmaParamBase object as property, so it gets pickled automatically
2021-11-18, cleanup comments
"""
import datetime as dt
import logging

import numpy as np

from EmaCalc.dirichlet import JEFFREYS_CONC
# = Jeffreys prior concentration for Dirichlet distribution

from EmaCalc.ema_base import EmaParamBase
from EmaCalc.ema_group import EmaGroupModel
from EmaCalc.ema_latent import Bradley


# -------------------------------------------------------------------
__ModelVersion__ = "2023-05-17"

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)  # *** TEST


# ------------------------------------------------------------------
class EmaModel:
    """Defines probability distributions of EMA recordings
    for all populations from which participants were recruited,
    and for all participants in each group,
    given all included EMA data,
    as collected in one ema_data.EmaDataSet instance.

    Model parameters define estimated probabilities for nominal SITUATIONS,
    and estimated distributions of latent variables that determine ordinal subjective ratings
    on given ATTRIBUTE questions.
    The latent variables have continuous values on an INTERVAL scale, as defined by the model,
    although all subjective Attribute ratings are discrete on ORDINAL scales.

    Each ATTRIBUTE latent-variable location may depend on one or more situation dimensions,
    as estimated by an ordinal regression model,
    including main and interaction effects as requested by the user.

    The model allows for the possibility that different respondents
    might interpret and use the ordinal rating scales in different ways,
    by including separate response-threshold parameters for each individual.

    The model allows for the possibility that each respondent
    might use different response thresholds for each Attribute question,
    even if the response alternatives are expressed by the same words.

    However, the researcher may also set up the ema_data.EmaFrame object such that
    the model presumes identical response thresholds for separate Attributes.
    """
    def __init__(self, base, groups, rng):
        """
        :param base: single common EmaParamBase object, used by all model parts
        :param groups: dict with EmaGroupModel instances, stored as
            groups[group] = one EmaGroupModel instance, where
                group = dict key = tuple of tuple(s) = (group_factor, factor_category)
            groups[group].participants[s_id] = an EmaRespondentModel instance, where
            s_id is a key identifying the participant.
        :param rng: random Generator object
        """
        self.base = base
        self.groups = groups
        self.rng = rng

    def __repr__(self):
        return self.__class__.__name__ + '(groups=groups)'

    @classmethod
    def initialize(cls, ds, effects,
                   max_n_comp=None,  # *** no longer used
                   n_participants_per_comp=5,
                   rv_class=Bradley,
                   restrict_attribute=False,
                   restrict_threshold=True,
                   seed=None):
        """Create a crude initial model from all available count-profile data.
        :param ds: a single ema_data.EmaDataSet instance with all EMA data for analysis
        :param effects: iterable with desired estimated effects of situatio on attribute attribute_grades.
            Each effect element = a key in ds.emf.situation_dtypes, or a tuple of such keys.
        :param max_n_comp: Not used, only for compatibility warning
        :param n_participants_per_comp: (optional) expected number participants per mixture component
            -> initial n_comp = n_participants // n_participants_per_comp
            The number of actually used components may be reduced during VI learning.
        :param rv_class: (optional) class of latent sensory random variable
        :param restrict_attribute: (optional) boolean switch
            to force restriction on attribute sensory-variable locations
        :param restrict_threshold: (optional) boolean switch
            to force restriction on response-threshold locations
        :param seed: (optional) integer to get reproducible random sequences
        :return: a cls instance
        """
        ds.ensure_complete()
        if max_n_comp is not None:
            logger.warning(f'Version >= 1.0: max_n_comp no longer used. Set n_participants_per_comp instead.')
        n_participants_per_comp = max(2, n_participants_per_comp)
        if restrict_attribute and ds.emf.tied_response_scales:
            restrict_attribute = False
            logger.warning('*** Cannot use restrict_attribute=True with tied response scales')
        if restrict_attribute and restrict_threshold:
            restrict_attribute = False  # ONLY ONE restriction allowed
            logger.warning(f'Only ONE restriction allowed: using restrict_threshold={restrict_threshold}')
        if not (restrict_attribute or restrict_threshold):
            logger.warning('Either restrict_attribute or restrict_threshold '
                           + 'should be True, to avoid artificial variance!')
        base = EmaParamBase.initialize(ds.emf, effects, rv_class,
                                       restrict_attribute=restrict_attribute,
                                       restrict_threshold=restrict_threshold)
        # = all base variables, to be used by all model objects
        seed_seq = np.random.SeedSequence(seed)
        rng = np.random.default_rng(seed_seq)
        # = main Generator for all random numbers, EXCEPT EmaRespondentModel instances
        # Same rng for all EmaGroupModel instances, but separate for EmaRespondentModel-s
        group_seeds = seed_seq.spawn(len(ds.groups))
        groups = {g: EmaGroupModel.initialize(n_participants_per_comp, base, g_data, g_seed, rng)
                  for ((g, g_data), g_seed) in zip(ds.groups.items(),
                                                   group_seeds)}
        logger.info('EmaModel initialized with ' +
                    f'{len(groups)} group(s); ' +
                    f'{base.n_parameters} model parameters;\n\t' +
                    f'n_participants_per_comp={n_participants_per_comp}; '
                    f'restrict_attribute = {restrict_attribute}; '
                    f'restrict_threshold = {restrict_threshold};')
        if seed is None:
            logger.debug(f'*** Using seed={seed}')
        else:
            logger.warning(f'*** Using seed={seed} -> reproducible results.')
        return cls(base, groups, rng)

    # ------------------------------------------ General VI learn algorithm:
    def learn(self,
              min_iter=10,
              min_step=0.1,
              max_iter=np.inf,
              max_hours=0.,
              max_minutes=0.,
              callback=None):
        """Learn all individual and population parameter distributions
        from all observed EMA data stored in self.groups[...].participants[...],
        using Variational Inference (VI).

        This method adapts a population GMM in each EmaGroupModel instance,
        and a sampled approximation of parameter distribution in all EmaRespondentModel instances.

        VI maximizes a lower bound to the total log-likelihood of all observed data.
        The resulting sequence of log-likelihood values is guaranteed to be non-decreasing,
        except for minor random variations caused by the sampling.

        :param min_iter: (optional) minimum number of learning iterations
        :param min_step: (optional) minimum data log-likelihood improvement,
                 over the latest min_iter iterations,
                 for learning iterations to continue.
        :param max_iter: (optional) maximum number of iterations, regardless of result.
        :param max_hours = (optional) maximal allowed running time, regardless of result.
        :param max_minutes = (optional) maximal allowed running time, regardless of result.
        :param callback: (optional) function to be called after each iteration step.
            If callable, called as callback(self, log_prob)
            where log_prob == scalar last achieved value of VI log-likelihood lower bound
        :return: log_prob = list of log-likelihood values, one for each iteration.

        Result: updated all properties of
        self.groups[...].pop_gmm properties comp, mix_weight
        self.groups[...].participants[...].xi = parameter sample array
        """
        if max_hours == 0. and max_minutes == 0.:
            max_minutes = 30.
        logger.info('Learning the model. Might take some time! '
                    + f'Max {max_hours:.0f} hours + {max_minutes:.0f} minutes.')
        min_iter = np.max([min_iter, 1])
        end_time = dt.datetime.now() + dt.timedelta(hours=max_hours,
                                                    minutes=max_minutes)
        # = last allowed time to start new VI iteration
        log_prob = []
        while (len(log_prob) <= min_iter
               or (log_prob[-1] - log_prob[-1 - min_iter] > min_step
                   and (len(log_prob) < max_iter)
                   and (dt.datetime.now() < end_time))):
            log_prob.append(self.adapt())
            if callable(callback):
                callback(self, log_prob[-1])
            logger.info(f'Done {len(log_prob)} iterations; data_LL={log_prob[-1]:.2f}')
        if dt.datetime.now() >= end_time:
            logger.warning('Learning stopped at time limit, possibly not yet converged')
        if len(log_prob) >= max_iter:
            logger.warning('Learning stopped at max iterations, possibly not yet converged')
        return log_prob

    def prune(self, min_weight=JEFFREYS_CONC):
        """Prune model to keep only active profile clusters
        :param min_weight: scalar, smallest accepted value for sum individual weight
        :return: None
        Result: all groups pruned with same criterion
        """
        for (g_name, g_model) in self.groups.items():
            # logger.info(f'Pruning group {g_name}:')
            g_model.prune(g_name, min_weight)

    def adapt(self):
        """One VI adaptation step for all groups
        :return: ll = scalar VI lower bound to data log-likelihood,
            incl. negative contributions for parameter KLdiv re priors
        """
        return sum(g_model.adapt(g_name)
                   for (g_name, g_model) in self.groups.items())


# ------------------------------------------------- TEST:
# if __name__ == '__main__':



"""This module implements Dirichlet-distributed random vectors,
similar to scipy.stats.dirichlet.

A Dirichlet-distributed vector is a random vector of fractions,
with sum == 1., and all elements in [0., 1.].

The Dirichlet-distributed vector may be used as a probability-mass parameter
for a corresponding Multinomial-distributed count array of integer elements,
or for the component weights in a Mixture Model.

*** Main Classes:
DirichletVector: model for Dirichlet-distributed probability-mass vector,
    with all elements in [0., 1.], and sum of all elements == 1.
    The distribution is specified by an array of concentration parameters.
    The concentration parameters can be adapted to observed data.

*** Module functions:
log_multi_beta = log normalization factor for Dirichlet distribution
d_log_multi_beta = gradient of log_multi_beta

*** Version History:
* Version 0.9.3:
2022-08-23, simplify old code, to allow only 1-dim random Dirichlet vector in EmaCalc version

* Earlier versions:
2019-12-22, first multidimensional version, including Multinomial
"""
import numpy as np
from scipy.special import gammaln, psi

import logging

# ------------------------------------------------------------------------
logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)  # test

JEFFREYS_CONC = 0.5


# --------------------------------------------------------------------------------
class DirichletVector:
    """Representing a multivariate Dirichlet-distributed vector.

    If U is a Dirichlet Random Vector, with elements U_i,
    the probability density function of U is determined by
    a vector of Concentration Parameters alpha, as
    p_U(u) = (1 / B(alpha) ) * prod_i u_i**(alpha_i - 1),
    B(alpha) = ( prod_i Gamma( alpha_i) ) / Gamma( sum_i alpha_i ),
    within the support domain
    0 < u_i < 1, all i, and sum_i( u_i ) == 1

    The normalization factor B(alpha) is implemented by module function
    log_multi_beta.
    """
    def __init__(self, alpha, rng=None):
        """
        :param alpha: 1-dim array-like sequence of concentration parameters
        :param rng: (optional) numpy.random.Generator object
        """
        self.alpha = np.array(alpha, dtype=float)
        if rng is None:
            self.rng = np.random.default_rng()
        else:
            self.rng = rng

    def __repr__(self):
        skip = '\n\t'
        return (self.__class__.__name__ + '(' + skip +
                (',' + skip).join(f'{key}={repr(v)}'
                                  for (key, v) in vars(self).items()) +
                skip + ')')

    # @classmethod
    # def initialize(cls, x, **kwargs):  # ********
    #     """Crude initial setting of concentration parameters
    #     :param x: array-like 1D list with non-normalized row vector(s)
    #         that might be generated from a cls instance,
    #         OR from a multinomial distribution with a cls instance as probability
    #         x[..., :] = ...-th row vector
    #     :param kwargs: (optional) any other key-word arguments for __init__
    #     :return: a new cls instance
    #     """
    #     a = np.array(x) + JEFFREYS_CONC
    #     # including Jeffreys concentration as pseudo-count
    #     a /= np.sum(a, axis=-1, keepdims=True)
    #     a *= JEFFREYS_CONC * a.shape[-1]
    #     # = normalized with average conc = JEFFREYS_CONC
    #     return cls(a, **kwargs)

    @property
    def shape(self):
        return self.alpha.shape

    @property
    def size(self):
        return self.alpha.size

    @property
    def ndim(self):  # == 1
        """Dimensionality of self"""
        return self.alpha.ndim

    @property
    def mean(self):
        return self.alpha / np.sum(self.alpha)

    @property
    def var(self):
        """Dirichlet variance Ref Bishop Eq B.18
        :return: v = array with
            v[...] = variance of ...-th element in random array
            v.shape == self.shape
        """
        a = self.alpha
        a0 = np.sum(a)  # across all elements regardless of shape
        return a * (a0 - a) / (a0**2 * (a0 + 1))

    @property
    def cov(self):
        """Ref Leijon PattRecKomp, Ch. 8
        Bishop Eq B.19 only for off-diag elements
        :return: c = 2D square covariance matrix for self, i.e.
            c[j, k] = covariance between (j,k)-th elements of self
            c.shape == (self.size, self.size)
        """
        return _cov_dirichlet(self.alpha)

    @property
    def mean_log(self):
        """E{ log U } where U is DirichletVector self
        :return: ml = vector with
            ml[k] = E{ log self[k] }
            ml.shape == self.shape
        """
        return psi(self.alpha) - psi(np.sum(self.alpha))

    # @property
    # def mean_root(self):  # *** not needed for EmaCalc
    #     """E{ sqrt(U) } element-wise, where U is DirichletVector self
    #     :return: ml = array with
    #         r[k] = E{ sqrt(self[k]) }
    #         r.shape == self.shape
    #     """
    #     a0 = np.sum(self.alpha)
    #     ln_r = (gammaln(a0) - gammaln(a0 + 0.5)
    #             + gammaln(self.alpha + 0.5) - gammaln(self.alpha))
    #     return np.exp(ln_r)

    # def mean_square_hellinger(self, othr):  # *** not needed in EmaCalc
    #     """Expected square Hellinger distance between self and othr,
    #     assuming independence
    #     :param othr: object with same class and shape as self
    #     :return: scalar h2 = E{H^2(self, othr)} = 1 - sum_k E{sqrt( u_k v_k ) }
    #         where u = (..., u_k, ...) = self, and v = (..., v_k, ...) = othr
    #     """
    #     return 1. - np.dot(self.mean_root, othr.mean_root)

    def relative_entropy(self, othr):
        """Relative Entropy = Kullback-Leibler divergence
        between two DirichletVector instances.
        :param othr: DirichletVector instance with same shape as self
        :return: KLdiv = scalar
            KLdiv = E{ log p_self(U) / p_othr(U) } evaluated for U = self
        """
        if self.shape == othr.shape:
            return (gammaln(np.sum(self.alpha)) - gammaln(np.sum(othr.alpha))
                    + np.sum(gammaln(othr.alpha) - gammaln(self.alpha))
                    + np.dot((self.alpha - othr.alpha).reshape(-1),
                             self.mean_log.reshape(-1))
                    )
        else:
            logger.warning('relative_entropy: Shape mismatch')
            return np.inf

    def logpdf(self, x):  # *** not needed for EmaCalc
        """log probability density of input sample vector(s).
        :param x: array-like (sequence of) vectors
            that might be samples drawn from self
        :return: lp = scalar or array with one value for each input sample array
            lp.shape == x.shape[:-1]
        """
        x = np.asarray(x)
        if x.ndim < 1:
            raise RuntimeError('Dimension mismatch')
        if x.shape[-1:] != self.shape:
            raise RuntimeError('Shape mismatch')
        # x and self.alpha are broadcast-compatible
        sum_axes = tuple(range(-1, 0))
        ok = np.logical_and(np.all(0. <= x, axis=-1),
                            np.isclose(1., np.sum(x, axis=-1)))
        if not np.all(ok):
            logger.warning('Some input is not in Dirichlet support -> logpdf = -inf')
        lp = (np.sum((self.alpha - 1.) * np.log(x), axis=sum_axes)
              - log_multi_beta(self.alpha))
        if np.isscalar(lp):
            if not ok:
                lp = -np.inf
        else:
            lp[np.logical_not(ok)] = -np.inf
        return lp

    def rvs(self, size=None):  # *** not needed for EmaCalc
        """Random sample array(s) drawn from self.
        :param size: (optional) scalar integer or tuple of ints with desired number of samples
        :return: u = array of random-generated probability-profile vectors
            Using numpy.random.Generation conventions:
            if size is None:
                u.shape == self.shape
            else:
                u.shape == (*size, *self.shape)
        """
        u = self.rng.dirichlet(alpha=self.alpha, size=size)
        if size is None:
            return u.reshape(self.shape)
        elif np.iterable(size):
            return u.reshape((*size, *self.shape))
        else:
            return u.reshape((size, *self.shape))


# --------------------------------------------------- general module functions
def log_multi_beta(a, axis=-1):
    """Log multivariate beta function
    = log normalization factor for Dirichlet distribution
    :param a: array-like concentration parameter ROW vector
        or array-like sequence of such row vectors
    :param axis: (optional) int or tuple of ints for axis to be reduced
    :return: log_multi_beta =  log-normalization value(s)
        log_multi_beta.shape == a.shape[:-1]
    """
    return (np.sum(gammaln(a), axis=axis, keepdims=False)
            - gammaln(np.sum(a, axis=axis, keepdims=False)))


# def d_log_multi_beta(a, axis=-1):
#     """Gradient of log_multi_beta
#     :param a: array-like sequence of concentration parameter vectors
#     :param axis: (optional) int or tuple of ints for axis to be summed
#     :return: d_log_multi_beta[..., k] = d log_multi_beta(a) / d a[..., k]
#         log_multi_beta.shape == a.shape
#     """
#     return psi(a) - psi(np.sum(a, axis=axis, keepdims=True))


# --------------------------------------------------- local module helper stuff
def _cov_dirichlet(a):
    """DirichletVector covariance, given concentration.
    Called recursively in case of multi-dimensional concentration array
    :param a: multi-dim array of concentration ROW vectors
    :return: c = array of square covariance matrices,
        one for each row vector in a.
        c.shape == (a.size, a.size)
    """
    # if a.ndim == 1:
    asum = np.sum(a)
    c = - a[:, np.newaxis] * a  # off-diagonal
    cd = a * (asum - a)  # diagonal
    cd_ind = np.diag_indices_from(c)
    c[cd_ind] = cd
    return c / (asum**2 * (asum + 1))
    # else:
    #     return np.array([_cov_dirichlet(a_i) for a_i in a])


# def _neg_logprob_conc(a, x, w, conc_prior):  # ****** local or static class method
#     """Objective function to optimize Dirichlet concentration parameters
#     using observed multinomial data
#     :param a: 1D array with tentative concentration parameters
#     :param x: 2D array with observed count profiles, externally shaped as
#         x[n, k] = k-th element of n-th observed count profile
#         x.shape[-1:] == a.shape
#     :param w: 1D array with weight factors in (0, 1.)
#         len(w) == x.shape[0] == (N,)
#     :param conc_prior: a ConcVector instance
#     :return: nlp = scalar = - w-weighted average of ln p(x_n | a),
#         un-normalized because x remains fixed during learning.
#     """
#     # a[k] = k-th element of a, for x[n, k]
#     # log_multi_beta(a).shape == ()
#     # log_multi_beta(a + x).shape == (len(x),)
#     lp = (np.dot(w, log_multi_beta(a) - log_multi_beta(a + x))
#           # - prior_logprob_conc(a))
#           - conc_prior.logprob(a))
#     # lp = scalar
#     return lp
#
#
# def _d_neg_logprob_conc(a, x, w, conc_prior):
#     """Gradient of _neg_logprob_conc w.r.t given vector of concentration parameters.
#     :param a: 1D array with tentative concentration parameters
#     :param x: 2D array with observed count profiles, externally shaped as
#         x[n, k] = k-th element of n-th observed count profile
#         x.shape[-1:] == a.shape
#     :param w: 1D array with weight factors in (0, 1.)
#         len(w) == x.shape[0] == (N,)
#     :param conc_prior: a ConcVector instance
#     :return: d_nlp = 1D array with
#         d_nlp[k] = d _neg_logprob_conc(a, x, w) / d_a[k]
#         d_nlp.shape == a.shape
#     """
#     # d_log_multi_beta(a).shape == a.shape
#     # d_log_multi_beta(a + x).shape == x.shape
#     dlp = (np.dot(w, d_log_multi_beta(a) - d_log_multi_beta(a + x))
#            - conc_prior.d_logprob(a))
#            # - d_prior_logprob_conc(a))
#     # dlp.shape == a.shape
#     return dlp


# --------------------------------------------------------------- TEST:
if __name__ == '__main__':

    from scipy.optimize import approx_fprime, check_grad
    from EmaCalc import ema_logging
    ema_logging.setup()

    len_conc = 4  # length of concentration vector(s)

    # --------------------------------------------
    print('*** Testing log_multi_beta behavior')
    nx = 5
    nu_prime = 0.5

    def log_l(a, nx, nu):
        """log prob multinomial as function of uniform conc and uniform obs
        :param a: tentative scalar conc value
        :param nx: scalar size of vector
        :param nu: scalar tentative prior param
        :return: log p(a | nx, nu) non-normalized
        """
        a_vec = np.ones(nx)
        a_vec[0] = a
        return (log_multi_beta(a_vec + nu)
                - log_multi_beta(a_vec))
    # --------------------------------------------------

    a = np.linspace(0.1, 5., 50)
    lp_a = np.array([log_l(a_i, nx, nu_prime) for a_i in a])
    print('   a= ', np.array2string(a, precision=3))
    print('lp_a= ', np.array2string(lp_a, precision=3))

    # --------------------------------------------
    # print('*** Testing gradient of log_multi_beta')
    #
    # # ----------------------------------------
    # def test_ln_B(a):
    #     return log_multi_beta(a)
    #
    # def test_d_ln_B(a):
    #     return d_log_multi_beta(a)
    # # -----------------------------------------
    #
    # test_conc = np.arange(len_conc) + 1.
    # print(f'test_conc = {test_conc}')
    # print(f'test_ln_B(test_conc) = {test_ln_B(test_conc)}')
    # print(f'test_ln_B([test_conc]) = {test_ln_B([test_conc])}')
    # print('test_d_ln_B =', test_d_ln_B(test_conc))
    #
    # err = check_grad(test_ln_B, test_d_ln_B, test_conc)
    # print('approx_grad = ', approx_fprime(test_conc,
    #                                       test_ln_B,
    #                                       epsilon=1e-6))
    # print('check_grad err = ', err)

    # --------------------------------------------
    print('\n*** Testing DirichletVector')
    from scipy.stats import dirichlet as scipy_dir

    test_conc = np.array([1., 2., 3., 4.] )
    # test_conc = np.ones(5)

    drv = DirichletVector(test_conc)
    print(f'drv = {drv}')
    print(f'drv.alpha = {drv.alpha}')
    print(f'drv.mean = {drv.mean}')
    if drv.ndim == 1:
        print(f'scipy_dir.mean = {scipy_dir(alpha=test_conc).mean()}')
    print(f'mean(drv.rvs(size=1000) = {np.mean(drv.rvs(size=1000), axis=0)}')

    # print(f'\ndrv.mean_root = {drv.mean_root}')
    # print(f'sqrt(drv.mean) = {np.sqrt(drv.mean)}')
    # print(f'mean(sqrt((drv.rvs(size=1000)) = {np.mean(np.sqrt(drv.rvs(size=1000)), axis=0)}')
    #
    # h2 = drv.mean_square_hellinger(drv)
    # print(f'\ndrv.mean_square_hellinger(drv) = {h2}')
    # r_u = np.sqrt(drv.rvs(size=1000))
    # r_v = np.sqrt(drv.rvs(size=1000))
    # print(f'mean_sample(1 - dot(sqrt(v),sqrt(v))) = {1. - np.mean(np.sum(r_u * r_v, axis=-1))}')
    # print(f'angle Hellinger= {np.arccos(1 - h2) * 180 / np.pi:.1f} degrees')

    print(f'\ndrv.var = {drv.var}')
    if drv.ndim == 1:
        print(f'scipy_dir.var = {scipy_dir(alpha=test_conc).var()}')

    print(f'var(drv.rvs(size=1000) = {np.var(drv.rvs(size=1000), axis=0)}')
    print(f'drv.cov = {drv.cov}')
    print(f'drv.mean_log = {drv.mean_log}')

    print(f'mean(log(drv.rvs(size=1000)) = {np.mean(np.log(drv.rvs(size=1000)), axis=0)}')
    x = drv.rvs()
    print(f'x = drv.rvs() = {x}')
    print(f'drv.logpdf(x) = {drv.logpdf(x)}')
    if drv.ndim == 1:
        print(f'scipy_dir.logpdf(x) = {scipy_dir.logpdf(x, alpha=test_conc)}')
    x = drv.rvs(size=1)
    print(f'x=drv.rvs(size=1) = {x}')
    print(f'drv.logpdf(x) = {drv.logpdf(x)}')
    # print(f'scipy_dir.logpdf(x) = {scipy_dir.logpdf(x, alpha=test_conc)}')  # not allowed
    x = drv.rvs(size=(2, 3))
    print(f'x = drv.rvs(size=(2, 3) = {x}')
    print(f'sum(x) = {np.sum(x, axis=-1)}')
    print(f'drv.logpdf(x) = {drv.logpdf(x)}')
    # print(f'scipy_dir.logpdf(x) = {scipy_dir.logpdf(x, alpha=test_conc)}')  # not allowed
    x[0, 0, 0] += 0.001  # test outside DirichletVector support
    print(f'sum(x) = {np.sum(x, axis=-1)}')
    print(f'drv.logpdf(x) = {drv.logpdf(x)}')



"""This module includes functions to format output displays of
EmaModel results, in graphic and textual form.

Plot properties may be controlled by
1: specifying matplotlib style sheet(s) by keyword argument mpl_style
2: setting specific matplotlib parameters at runtime by keyword argument mpl_params
3: setting specific parameters in FMT, e.g., 'colors' and 'markers'

*** Version History:
* Version 0.9.6:
2023-04-16, some plot parameters -> FMT for user control
2023-04-12, new simplified tab_percentles, tab_credible_diff
2023-03-31, new help function make_product_name *** NO LONGER NEEDED

* Version 0.9.5:
2023-03-07, fig_category_barplot allow both observed and model-predicted quantile data

* Version 0.9.4:
2023-01-22, Bug fix in output file name creation, to avoid problem under Windows.
            Added 'interaction_sep' and 'condition_sep' in FMT dict, for user control.
            Added try...catch for any errors in ResultPlot.save() and ResultTable.save()

* Version 0.9.3:
2022-07-27, allow setting matplotlib style sheet(s) and individual matplotlib params
2022-07-12, Special DiffTable class to suppress numerical index in save.
2022-07-12, Simplified header in tab_credible_diff.

* Version 0.9.2:
2022-06-17, fig_percentiles: allow caller to set y_min, y_max
2022-06-04, suppress integer index column in Table.save. New support function harmonize_ylim
2022-06-03, tab_credible_diff with clarified header

* Version 0.9.1:
2022-04-10, fig_percentiles using DataFrame table as input
2022-03-30, make all tables as pandas DataFrame objects

* Version 0.8:
2022-02-15, minor cleanup of tab_percentile: allow cdf=None, header percentile format .1f

* Version 0.7.1:
2022-01-21, minor fix in tab_cred_diff to avoid credibility == 100.0%
2022-01-08, set_tight_layout for all ResultPlot objects
2022-01-13, changed EmaDisplaySet.show format argument: show_intervals -> grade_thresholds

* Version 0.7:
2021-12-19, function nap_table to format NAP results

2021-11-07, copied and modified PairedCompCalc -> EmaCalc
"""
import numpy as np
from itertools import cycle
import matplotlib.pyplot as plt
import logging
import pandas as pd

from .ema_file import Table

logger = logging.getLogger(__name__)


FMT = {'colors': 'rbgk',    # to distinguish results in plots, cyclic use
       'markers': 'oxs*_',  # corresponding markers, cyclic use
       'x_space': 0.3,      # clean space between x_tick categories
       'interaction_sep': '\u00D7',  # mult.sign. separating situation labels in result file names
       'condition_sep': '_',    # separating attribute and its conditioning situation(s) in file names
       'max_plot_cases': 50,    # max cases in percentile plots
       'max_case_heads': 1,     # limit in plot legends
       'and_head': ('', ''),    # two-level heading for diff-table first column
       'and_label': 'and',      # logical and in diff-table first column
       }

# NOTE: FMT['colors'] and FMT['markers'] override matplotlib.rcParams.axes.prop_cycle,
#   because prop_cycle allows only equal lengths of 'colors' and 'markers'.
#   The FMT['colors'] and FMT['markers'] are used cyclically,
#   so the default sequences with unequal lengths will combine
#   into a sequence with many combinations, before repeating itself.


def set_format_param(mpl_style=None, mpl_params=None, **kwargs):
    """Set / modify format parameters.
    Called before any displays are generated.
    :param mpl_style: (optional) matplotlib style sheet, or list of style sheets
    :param mpl_params: (optional) dict with matplotlib (k, v) rcParam settings
    :param kwargs: dict with any formatting variables to be stored in FMT
    :return: None
    """
    if mpl_style is not None:
        plt.style.use(mpl_style)
    if mpl_params is not None:
        plt.rcParams.update(mpl_params)
    other_fmt = dict()
    for (k, v) in kwargs.items():
        k = k.lower()
        if k in FMT:
            FMT[k] = v
        else:
            other_fmt[k] = v
    if len(other_fmt) > 0:
        logger.warning(f'Parameters {other_fmt} unknown, not used.')


# ---------------------------- Basic Result Classes

class ResultPlot:
    """Container for a single graph instance
    """
    def __init__(self, ax, name):
        """
        :param ax: matplotlib Axes instance containing the graph
        :param name: string identifying the plot; used as file name
        """
        self.ax = ax
        self.name = name

    @property
    def fig(self):
        return self.ax.figure

    def save(self, path,
             figure_format,
             **kwargs):
        """Save figure to given path
        :param path: Path to directory for saving self
        :param figure_format: figure-format string code -> file-name suffix
        :param kwargs (optional) any additional kwargs, *** NOT USED ***
        :return: None
        """
        # *** select subset of kwargs allowed by savefig() ?
        # NO, depends on Matplotlib backend!
        f = (path / self.name).with_suffix('.' + figure_format)
        try:
            self.fig.savefig(f)
        except Exception as e:  # any error, just warn and continue
            logger.warning(f'Could not save plot to {f}. Error: {e}')


class ResultTable(Table):
    """A pd.DataFrame table subclass, with a name and special save method
    """
    def __init__(self, df, name):
        """
        :param df: a Table(pd.DataFrame) instance
        :param name: file name for saving the table
        """
        super().__init__(df)
        self.name = name

    def save(self, path,
             table_format='txt',
             **kwargs):
        """Save table to file.
        :param path: Path to directory for saving self.
            suffix is determined by FMT['table_format'] anyway
        :param table_format: table-format string code -> file-name suffix
        :param kwargs: (optional) any additional arguments to pandas writer function
        :return: None
        """
        f = (path / self.name).with_suffix('.' + table_format)
        try:
            super().save(f, **kwargs)
        except Exception as e:  # any error, just warn and continue
            logger.warning(f'Could not save result table. Error: {e}')


class DiffTable(ResultTable):
    """Special subclass suppressing index in save method
    """
    def save(self, path,
             **kwargs):
        """Save table to file.
        :param path: Path to directory for saving self.
            suffix is determined by FMT['table_format'] anyway
        :param table_format: table-format string code -> file-name suffix
        :param kwargs: (optional) any additional arguments to pandas writer function
        :return: None
        """
        if 'index' not in kwargs:
            kwargs['index'] = False  # override Pandas default = True
        super().save(path, **kwargs)


# ---------------------------------------- Formatting functions:

def fig_percentiles(df,
                    y_label='',
                    file_label='',
                    cat_limits=None,
                    y_min=None,
                    y_max=None,
                    **kwargs):
    """create a figure with percentile results
    as defined in a given pd.DataFrame instance
    :param df: pd.DataFrame instance with primary percentile data, with
        one row for each case category, as defined in df.index.values elements,
        one column for each percentile value.
    :param x_label: (optional) string for x-axis label, instead of df.index.names[0]
    :param x_tick_labels: (optional) list of x tick labels, instead of df.index.levels[0]
    :param y_label: (optional) string for y-axis label
    :param y_label: (optional) string for y-axis label
    :param cat_limits: 1D array with response-interval limits (medians)
    :param file_label: (optional) string as first part of file name
    :param colors: (optional) sequence of color codes to separate results in plots
    :param markers: (optional) sequence of marker codes to separate results in plots
        len(colors) != len(markers) -> many different combinations
    :param y_min: (optional) enforced lower limit of vertical axis
    :param y_max: (optional) enforced upper limit of vertical axis
    :param kwargs: (optional) dict with any additional keyword arguments for plot commands.
    :return: ResultPlot instance with plot axis with all results
    NOTE: plot will use df.index.level[0] categories as x-axis labels,
    and index.level[1:] as plot labels in the legend
    """
    # ----------------------------------- set up plot design:
    if df is None:
        return None
    if df.index.nlevels == 1:
        x_label = df.index.name
        x_tick_labels = list(df.index.values)
        case_head = ()
        case_list = [()]
    elif df.index.nlevels == 2:
        x_label = df.index.names[0]
        x_tick_labels = list(dict.fromkeys([c[0] for c in df.index.values]))
        case_head = df.index.names[1]
        case_list = list(dict.fromkeys([c[1] for c in df.index.values]))
    else:
        x_label = df.index.names[0]
        x_tick_labels = list(dict.fromkeys([c[0] for c in df.index.values]))
        case_head = tuple(df.index.names[1:])
        case_list = list(dict.fromkeys([c[1:] for c in df.index.values]))
        # in order as appearing in df
    if df.shape[0] > FMT['max_plot_cases']:
        logger.warning(f'Too many {x_label}, {case_head} cases to plot. Increase max_plot_cases if necessary.')
        return None
    n_cases = len(case_list)
    dx = (1. - FMT['x_space']) / n_cases
    # = x step between range plots for separate cases
    x_offset = {c: (i - (n_cases - 1) / 2) * dx
                for (i, c) in enumerate(case_list)}
    case_color = {c: col for (c, col) in zip(case_list, cycle(FMT['colors']))}
    case_marker = {c: mark for (c, mark) in zip(case_list, cycle(FMT['markers']))}
    # ------------------------------------------------------------------------
    def plot_one(ax, row_index, y):
        """
        :param ax: axis for plot
        :param row_index: index label(s) for ONE row in df
        :param y: 1D array with corresponding y = values in ONE row
            y[p] = p-th percentile value
        :return: None
        """
        if df.index.nlevels == 1:
            x_tick = row_ind
            x_case = ()
        elif df.index.nlevels == 2:
            x_tick = row_ind[0]
            x_case = row_ind[1]
        else:
            x_tick = row_ind[0]
            x_case = row_ind[1:]
        x = x_tick_labels.index(x_tick) + x_offset[x_case]
        x = x * np.ones_like(y)
        c = case_color[x_case]
        m = case_marker[x_case]
        if len(y) == 1:  # only single marker
            line = ax.plot(x, y,
                           linestyle='', color=c,
                           marker=m, markeredgecolor=c, markerfacecolor='w',
                           **kwargs)
        elif len(y) == 2:  # vertical range, no markers
            line = ax.plot(x, y,
                           linestyle='solid', color=c,
                           marker=m, markeredgecolor=c, markerfacecolor='w',
                           **kwargs)
        else:  # vertical range, and markers for intermediate percentiles
            y = sorted(list(y))
            ax.plot([x[0], x[-1]], [y[0], y[-1]],
                    linestyle='solid', color=c,
                    **kwargs)
            line = ax.plot(x[1:-1], y[1:-1],
                           linestyle='solid', color=c,  # to get line+marker into label
                           marker=m, markeredgecolor=c, markerfacecolor='w',
                           **kwargs)
        if x_tick == x_tick_labels[0]:
            if type(case_head) is tuple and len(case_head) > FMT['max_case_heads']:
                line[0].set_label(str(x_case))
            else:
                line[0].set_label(str(case_head) + '=' + str(x_case))

    # ------------------------------------------------------------------
    fig, ax = plt.subplots()
    for (row_ind, y) in df.iterrows():
        plot_one(ax, row_ind, y)

    (x_min, x_max) = ax.get_xlim()
    x_min = min(x_min, -FMT['x_space'])
    x_max = max(x_max, len(x_tick_labels) - 1 + FMT['x_space'])
    ax.set_xlim(x_min, x_max)
    if cat_limits is not None:
        _plot_response_intervals(ax, cat_limits)
    ax.set_xticks(np.arange(len(x_tick_labels)))
    xticks = [str(c) for c in x_tick_labels]
    ax.set_xticklabels(xticks,
                       **_x_tick_style(xticks))
    (y0, y1) = ax.get_ylim()
    if y_min is not None:
        y0 = y_min
    if y_max is not None:
        y1 = y_max
    ax.set_ylim(y0, y1)
    ax.set_ylabel(y_label)
    ax.set_xlabel(x_label)
    if len(case_list) > 1:
        ax.legend(loc='best')
    if len(file_label) > 0:
        file_label += FMT['condition_sep']
    f_name = file_label + FMT['interaction_sep'].join(df.index.names)
    fig.set_tight_layout(tight=True)  # **** let rcParams control this ? ****
    return ResultPlot(ax, name=f_name)


def _plot_response_intervals(ax, c_lim):
    """plot horizontal lines to indicate response-category intervals
    :param ax: axis object
    :param c_lim: 1D array with scalar interval limits
    :return: None
    """
    (x_min, x_max) = ax.get_xlim()
    return ax.hlines(c_lim, x_min, x_max,
                     colors=plt.rcParams['axes.edgecolor'],
                     linewidth=0.2 * plt.rcParams['lines.linewidth'])


def fig_category_barplot(df,
                         x_label,
                         y_label,
                         df_q=None,
                         file_label='',
                         y_min=None,  # *** not needed, always = 0
                         y_max=None,  # *** not needed, adaptive
                         mpl_params=None,  # *** not needed? called only from ema_display ***
                         **kwargs
                         ):
    """Bar plot of DataFrame values,
    to be displayed with one sequence of vertical bars along x-axis for each row,
    with one bar for each column,
    suitable, e.g., for plotting attribute_grade_counts
    :param df: a DataFrame instance,
        one row for each selected situation category, one column for each grade
    :param df_q: (optional) DataFrame instance with quantiles
        similar to df, but expanded with one row for each (situation category, quantile)
    :param x_label: x-axis label string
    :param y_label: y-axis label string
    :param file_label: plot name for saving file
    :param y_min: (optional) enforced lower limit of vertical axis *** NOT USED
    :param y_max: (optional) enforced upper limit of vertical axis *** NOT USED
    :param mpl_param: (optional) dict with matplotlib rcParam settings
    :param kwargs: (optional) dict with keyword arguments for plot commands *** NOT USED
    :return: a ResultPlot instance
    """
    if mpl_params is not None:
        plt.rcParams.update(mpl_params)
    fig, ax = plt.subplots()
    assert df.ndim > 1, 'Input must be DataFrame'
    (n_cases, n_x) = df.shape
    x = np.arange(n_x)
    bar_space = 0.02  # space to allow all bar edges to be visible
    w = 0.8 / n_cases - bar_space
    if len(df.index.names) > 1:
        case_head = tuple(df.index.names)
    else:
        case_head = df.index.name
    x_dev = (w + bar_space) * (np.arange(n_cases) - (n_cases - 1) / 2)
    for (d, c, case) in zip(x_dev,
                            cycle(FMT['colors']),
                            df.index.values):
        y = df.loc[case].to_numpy()
        # *** check case_head vs FMT['max_case_head'] ***
        if type(case_head) is tuple and len(case_head) > FMT['max_case_heads']:
            case_legend = str(case)
        else:
            case_legend = str(case_head) + '= ' + str(case)
        ax.bar(x + d, height=y,
               width=w, edgecolor=c, facecolor='w',  # ************
               label=case_legend)
    if df_q is not None:
        for (d, c, case) in zip(x_dev,
                                cycle(FMT['colors']),
                                df.index.values):
            y = df_q.loc[case].to_numpy()
            ax.plot(np.tile(x + d, (len(y),1)), y, '-', color=c, linewidth=2)
    ax.set_xticks(np.arange(n_x))
    xticks = [str(c) for c in df.columns.values]
    ax.set_xticklabels(xticks,
                       **_x_tick_style(xticks))
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)
    if n_cases > 1:
        ax.legend(loc='best')
        f_name = file_label + FMT['condition_sep'] + FMT['interaction_sep'].join(df.index.names)
    else:
        f_name = file_label
    fig.set_tight_layout(tight=True)
    return ResultPlot(ax, name=f_name)


# ----------------------------------------- table displays:

def tab_percentiles(q_ds, file_label=''):
    """Re-format a pd.DataFrame with percentile results.
    :param q_ds: a pd.Series instance with quantiles as last index axis,
        and cases in other index axes.
    :param file_label: (optional) string as first part of file name
    :return: a ResultTable(pd.DataFrame) instance,
        with one column for each percentile,
        and one row for each combination product of cases considered.
    """
    tab_perc = q_ds.unstack()
    case_head = tab_perc.index.names
    tab_perc.columns = [f'{q:.1%}' for q in tab_perc.columns]
    if len(file_label) > 0:
        file_label += FMT['condition_sep']
    f_name = file_label + FMT['interaction_sep'].join(case_head)
    return ResultTable(tab_perc, name=f_name)


def tab_credible_diff(diff,
                      diff_head,
                      cred_head,
                      case_head=(),
                      y_label='',
                      file_label='',
                      and_label=FMT['and_label'],
                      and_head=FMT['and_head']   # ('', '')  # -> FMT ***
                      ):
    """Create table with credible differences among results -- pandas version
    :param diff: list of tuples (((i,j), c0,...), p),
        defining jointly credible differences, indicating that
        prob{ quality of diff category i > quality of category j, given case category c
        AND all previous pairs } == p
        i, j are either a string label or a tuple of such labels
        c0,... is one or more case labels
    :param diff_head: tuple of keys for heading of diff_labels column in table
        len(diff_head) == len(i) == len(j) if tuple, or len(diff_head) == 1
    :param cred_head: string for header of Credibility column
    :param case_head: (optional) tuple of case keys, one for each case-dimension table column
        len(case_head) == len(c0,...)
    :param y_label: (optional) string with label of tabulated attribute
    :param file_label: (optional) string for first part of file name
    :param and_label: (optional) joining AND label in first column
    :param and_head: (optional) tuple with two strings for head first column
    :return: ResultTable object with header lines + one line for each credible difference
    """
    if len(diff) == 0:
        return None
    y_head_i = y_label + ' >'
    y_head_j = y_label
    # --------------------- table columns as dicts:
    col = {and_head:  [' '] + [and_label] * (len(diff) - 1)}  # first column with only AND flags
    diff_i = [(ijc[0][0] if type(ijc[0][0]) is tuple else (ijc[0][0],)) for (ijc, p) in diff]
    diff_j = [(ijc[0][1] if type(ijc[0][1]) is tuple else (ijc[0][1],)) for (ijc, p) in diff]
    diff_c = [ijc[1:] for (ijc, p) in diff]
    diff_p = [p for (ijc, p) in diff]
    # --------- column(s) for higher results:
    col |= {(y_head_i, d_head_k): [d[k] for d in diff_i]
            for (k, d_head_k) in enumerate(diff_head)}
    # --------- column(s) for lower results:
    col |= {(y_head_j, d_head_k): [d[k] for d in diff_j]
            for (k, d_head_k) in enumerate(diff_head)}  # cols for lesser results
    # --------- column(s) for optional case labels:
    if len(case_head) > 0:
        # diff_c = [case_labels[d[0][2]]
        #           for d in diff]
        col |= {('', c_head_k): [d[k] for d in diff_c]
                for (k, c_head_k) in enumerate(case_head)}
    # --------- credibility column:
    col |= {('', cred_head): diff_p}
    df = pd.DataFrame(col)  # do all this in samppy.credibility_pd ? ***********
    # each column name is a tuple with two elements -> MultiIndex with two levels
    # df = df.reindex(columns=pd.MultiIndex.from_tuples(df.columns))  # ****** Needed ?
    if len(file_label) > 0:
        file_label += FMT['condition_sep']
    f_name = file_label + FMT['interaction_sep'].join(diff_head) + '-diff'
    if len(case_head) > 0:
        f_name += FMT['condition_sep'] + FMT['interaction_sep'].join(case_head)
    return DiffTable(df, name=f_name)


# -------------------------------------- display adjustment functions
def harmonize_ylim(axes_list, y_min=None, y_max=None):
    """Adjust several plots to equal vertical range
    :param axes_list: sequence of plt.Axes instances
    :param y_min: (optional) extra user-defined minimum
    :param y_max: (optional) extra user-defined maximum
    :return: None
    """
    y0 = min(*(ax.get_ylim()[0]
               for ax in axes_list))
    if y_min is not None:
        y0 = min(y0, y_min)
    y1 = max(*(ax.get_ylim()[1]
               for ax in axes_list))
    if y_max is not None:
        y1 = max(y1, y_max)
    for ax in axes_list:
        ax.set_ylim(y0, y1)


# -------------------------------------- private help functions
def _x_tick_style(labels):
    """Select xtick properties to avoid tick-label clutter
    :param labels: list of tick label strings
    :return: dict with keyword arguments for set_xticklabels
    """
    maxL = max(len(l) for l in labels)
    rotate_x_label = maxL * len(labels) > 75  # ad hoc criterion
    if rotate_x_label:
        style = dict(rotation=15, horizontalalignment='right')
    else:
        style = dict(rotation='horizontal', horizontalalignment='center')
    return style



"""This module defines help classes to calculate response thresholds
from given model parameters.

*** Classes:
Thresholds: OLD model version allowing all response thresholds freely adaptable.
ThresholdsFree: allowing all response thresholds freely adaptable, new version.
ThresholdsMidFixed: forcing one mid-range threshold -> zero, other thresholds free

*** Version history:
* Version 0.9.5: NEW module, with functions moved from ema_base
"""
import numpy as np
from scipy.special import logit, expit
# ***** logit, expit are NOT symmetric around mid-point for extreme arguments
# logit(expit(37.)) = 37.; logit(expit(38.)) = +inf; logit(expit(-38.)) = -38.
# logit(expit(-709.)) = -709.; logit(expit(-710.)) = -inf
# ***** need safer variant here? can use scipy.special.log_expit?
# ***** No, sufficiently protected by mapped_width() function

from scipy.special import logsumexp, softmax

ETA_W_EPSILON = np.finfo(float).eps
# = additive constant to prevent too small mapped_width(eta)


# -------------------------------------------------------------
class Thresholds:
    """Superclass for threshold calculations, given parameters,
    using the simple approach with one log-width parameter eta
    for each response interval / response category
    i.e., M free parameters but only M-1 free model thresholds.
    """
    @staticmethod
    def n_param(n_categories):
        """Number of log-category-width parameters, given number of response categories
        :param n_categories: integer number of response categories
                == number of response intervals
        :return: integer number of log-category-width parameters
                needed to specify response thresholds
        """
        return n_categories

    @staticmethod
    def tau(eta):
        """Mapping of given log-category-width parameters to response thresholds.
        :param eta: 1D or 2D array with
            eta[..., m] = ...-th sample of parameter defining
                non-normalized log-width of m-th interval in mapped domain [0, 1].
            eta.shape[-1] == self.n_param(M), with M == number of response categories.
        :return: tau = 1D or 2D array, incl. all elements in [-inf, +inf]
            (tau[..., m], tau[..., m+1] = (LOWER, UPPER) limits for m-th ordinal response interval
            tau[..., 0] ==  - np.inf
            tau[..., -1] == + np.inf
            tau.ndim == eta.ndim; tau.shape[-1] == M + 1, with M == number of response categories.
        """
        cum_w = np.cumsum(mapped_width(eta), axis=-1)
        z_shape = (*cum_w.shape[:-1], 1)
        cum_w = np.concatenate((np.zeros(z_shape), cum_w),
                               axis=-1)  # include cum_w[..., 0] = 0.
        # sum_w = cum_w[..., -1:]
        return logit(cum_w / cum_w[..., -1:])

    @staticmethod
    def d_tau(eta):
        """Jacobian of tau(eta) with respect to eta
        :param eta: = 1D or 2D array with
            eta[..., m] = ...-th sample of parameter defining
                non-normalized width of m-th interval in mapped domain [0, 1].
            eta.shape[-1] == self.n_param(M), with M == number of response categories.
        :return: 2D or 3D array d_tau, with
            d_tau[..., m, i] = d tau[..., m] / d eta[..., i],
                for m = 0,..., M + 1; i = 0, ..., self.n_param(M) - 1
                where (tau[s, m], tau[s, m+1]) = (LOWER, UPPER) limits of m-th response interval
            d_tau[..., 0, :] = d_tau[..., -1, :] == 0., for extreme limits at +-inf
            d_tau.shape == (N, M+1, self.n_param(M)); N = n_samples; M = n_categories
        """
        w = mapped_width(eta)
        # (n_samples, nw) = w.shape
        nw = w.shape[-1]
        cum_w = np.cumsum(w, axis=-1)
        cw = cum_w[..., :-1, np.newaxis]  # only inner limits
        sw = cum_w[..., -1:, np.newaxis]
        # tau[..., m+1] = ln cw[..., m]  - ln (sw[..., 0] - cw[..., m])
        # dcw_dw[..., m, i] = dcw[..., m] / dw[..., i]  = 1. if i <= m else 0.
        dcw_dw = np.tril(np.ones((nw - 1, nw), dtype=int))
        dtau_dw = dcw_dw / cw - (1 - dcw_dw) / (sw - cw)
        dtau_d_eta = dtau_dw * d_mapped_width_d_eta(w)[..., np.newaxis, :]
        # dtau_d_eta.shape = (n_samples, nw - 1, nw)  OR (nw - 1, nw)
        z_shape = (*dtau_d_eta.shape[:-2], 1, dtau_d_eta.shape[-1])
        return np.concatenate((np.zeros(z_shape),  # *** use np.pad ???
                               dtau_d_eta,
                               np.zeros(z_shape)), axis=-2)

    @staticmethod
    def tau_inv(tau):
        """Inverse of tau(eta)
        :param tau: 1D or 2D array with response thresholds, EXCEPT extremes at +-inf,
            i.e., all tau elements in (-inf, +inf),
            tau[..., m] = UPPER limit for m-th interval,
                = LOWER limit for the (m+1)-th interval
            tau.shape[-1] == M - 1 == number of response intervals - 1
        :return: eta: 1D or 2D array with
            eta[..., m] = ...-th sample of log non-normalized width of m-th interval.
            eta.shape[-1] == M == number of response intervals.

        Method:
            Normalized widths and interval limits are defined in transformed domain (0, 1.),
            using a logistic mapping function,
            y = expit(tau), where y in (0, 1]
                y[..., m] =  (w_0 +...+ w_m) / (w_0 + ... + w_{M-1};  0 <= m <= M-1
                w_m = _w(eta[..., m])
        """
        y = expit(tau)
        cat_shape = (*y.shape[:-1], 1)
        # include extreme limits at 0 and 1:
        y = np.concatenate((np.zeros(cat_shape),
                            y,
                            np.ones(cat_shape)), axis=-1)
        w = np.diff(y, axis=-1)
        return mapped_width_inv(w)


ThresholdsOld = Thresholds


# -------------------------------------------------------------
class ThresholdsFree(Thresholds):
    """All internal thresholds freely variable, as specified by model parameter vector eta,
    with number of parameters == number of internal thresholds == M - 1, where
    M = number of response categories
    """
    @staticmethod
    def n_param(n_categories):
        return n_categories - 1

    @staticmethod
    def tau(eta):
        """Mapping given log-category-width parameters to response thresholds.
        :param eta: = 1D or 2D array with
            eta[..., m] = ...-th sample of parameter defining
                non-normalized width of m-th interval in mapped domain [0, 1].
            eta.shape[-1] == self.n_param(M), with M == number of response categories.
            eta[..., M] assumed fixed == 0, NOT included in input argument
        :return: tau = 1D or 2D array, incl. all elements in [-inf, +inf]
            (tau[..., m], tau[..., m+1]) = (LOWER, UPPER) limits for m-th ordinal response interval
            tau[..., 0] ==  - np.inf
            tau[..., -1] == + np.inf
            tau.ndim == eta.ndim; tau.shape[-1] == eta.shape[-1] + 2
        """
        z_shape = (*eta.shape[:-1], 1)
        eta = np.concatenate((eta, np.zeros(z_shape)), axis=-1)
        return Thresholds.tau(eta)

    @staticmethod
    def d_tau(eta):
        """Jacobian of thresholds with respect to eta
        :param eta: = 1D or 2D array with
            eta[..., m] = ...-th sample of parameter defining
                 non-normalized width of m-th interval in mapped domain [0, 1].
            eta.shape[-1] == M - 1, where M == number of response intervals.
            eta[..., M] assumed fixed == 0, NOT given as input
        :return: 2D or 3D array d_tau, with
            d_tau[..., m, i] = d tau[..., m] / d eta[..., i]; m = 0,..., M; i = 0, ..., M-1
                 where (tau[s, m], tau[s, m+1] = (LOWER, UPPER) limits of m-th response interval
            d_tau[..., 0, :] = d_tau[..., -1, :] = 0., for extreme limits at +-inf
            d_tau.ndim == eta.ndim + 1; d_tau.shape[-2:] == (M+1, M-1)
        """
        z_shape = (*eta.shape[:-1], 1)
        zeta = np.concatenate((eta, np.zeros(z_shape)), axis=-1)
        d_tau_d_zeta = Thresholds.d_tau(zeta)
        return d_tau_d_zeta[..., :, :-1]

    @staticmethod
    def tau_inv(tau):
        """Inverse of tau(eta)
        :param tau: 1D or 2D array with response thresholds, EXCEPT extremes at +-inf,
            i.e., all tau elements in (-inf, +inf),
            tau[..., m] = UPPER limit for m-th interval,
                = LOWER limit for the (m+1)-th interval
            tau.shape[-1] == M - 1; M == number of response intervals
        :return: eta: 1D or 2D array with
            eta[..., m] = ...-th sample of log non-normalized width of m-th interval.
            eta[..., M] == 0. NOT included
            eta.shape[-1] == self.n_param(M) for M == number of response intervals.
        """
        eta = Thresholds.tau_inv(tau)
        # last element always -> 0., NOT included:
        return eta[..., :-1] - eta[..., -1:]


# -------------------------------------------------------------
class ThresholdsMidFixed(Thresholds):
    """One mid-range threshold fixed == 0.,
    other thresholds mapped from parameter array eta,
    with number of parameters == M - 2, where
    M = number of response categories
    """
    @staticmethod
    def n_param(n_categories):
        return n_categories - 2

    @staticmethod
    def tau(eta):
        """Mapping given log-category-width parameters to response thresholds.
        :param eta: = 1D or 2D array with
            eta[..., m - 1] = ...-th sample of parameter defining
                non-normalized width of m-th interval in mapped domain [0, 1].
            eta.shape[-1] == self.n_param(M), with M == number of response categories.
            w[..., 0] assumed fixed == 1, NOT given as input
            w[..., M] assumed fixed == 1, NOT given as input
        :return: tau = 1D or 2D array, incl. all elements in [-inf, +inf]
            (tau[..., m], tau[..., m+1]) = (LOWER, UPPER) limits for m-th ordinal response interval
            tau[..., 0] ==  - np.inf
            tau[..., -1] == + np.inf
            tau.ndim == eta.ndim; tau.shape[-1] == eta.shape[-1] + 2
        """
        z_shape = (*eta.shape[:-1], 1)
        zeta = np.concatenate((np.zeros(z_shape), eta, np.zeros(z_shape)),
                              axis=-1)
        n_half = zeta.shape[-1] // 2
        zeta[..., :n_half] -= logsumexp(zeta[..., :n_half], axis=-1, keepdims=True)
        zeta[..., n_half:] -= logsumexp(zeta[..., n_half:], axis=-1, keepdims=True)
        # = normalized to same total width in both halves
        return Thresholds.tau(zeta)
        # **** old method:
        # z_shape = (*eta.shape[:-1], 1)
        # v = np.concatenate((np.ones(z_shape), mapped_width(eta), np.ones(z_shape)),
        #                    axis=-1)
        # # = non-normalized widths
        # n_half = v.shape[-1] // 2
        # # w[..., :n_half] /= np.sum(w[..., :n_half], axis=-1, keepdims=True)
        # # w[..., n_half:] /= np.sum(w[..., n_half:], axis=-1, keepdims=True)
        # w = np.concatenate((v[..., :n_half] / np.sum(v[..., :n_half], axis=-1, keepdims=True),
        #                     v[..., n_half:] / np.sum(v[..., n_half:], axis=-1, keepdims=True)),
        #                    axis=-1)
        # # -> lower and upper intervals normalized separately to equal sum
        # cum_w = np.concatenate((np.zeros(z_shape), np.cumsum(w, axis=-1)),
        #                        axis=-1)
        # # sum_w = cum_w[..., -1:]
        # tau = logit(cum_w / cum_w[..., -1:])
        # if not np.all(np.isclose(tau, tau_new)):
        #     err_max = np.amax(tau[..., 1:-1] - tau_new[..., 1:-1], axis=0)
        #     err_min = np.amin(tau[..., 1:-1] - tau_new[..., 1:-1], axis=0)
        #     print('tau != tau_new')
        # return tau

    @staticmethod
    def d_tau(eta):
        """Jacobian of tau(eta) w.r.t. eta
        :param eta: = 1D or 2D array with
            eta[..., m - 1] = ...-th sample of parameter defining
                non-normalized width of m-th interval in mapped domain [0, 1].
            eta.shape[-1] == self.n_param(M), with M == number of response categories.
        :return: 2D or 3D array d_tau, with
            d_tau[..., m, i] = d tau[..., m] / d eta[..., i]; m = 0,..., M; i = 0, ..., M-1
                 where (tau[s, m], tau[s, m+1] = (LOWER, UPPER) limits of m-th response interval
            d_tau[..., 0, :] = d_tau[..., -1, :] = 0., for extreme limits at +-inf
            d_tau.ndim == eta.ndim + 1; d_tau.shape[-2:] == (M+1, M-2)
        """
        z_shape = (*eta.shape[:-1], 1)
        zeta = np.concatenate((np.zeros(z_shape), eta, np.zeros(z_shape)),
                              axis=-1)
        n_half = zeta.shape[-1] // 2
        zeta[..., :n_half] -= logsumexp(zeta[..., :n_half], axis=-1, keepdims=True)
        zeta[..., n_half:] -= logsumexp(zeta[..., n_half:], axis=-1, keepdims=True)
        # = normalized to same total width in both halves
        d_zeta_d_eta = np.tile(np.identity(zeta.shape[-1]),
                               (*zeta.shape[:-1], 1, 1))
        d_zeta_d_eta[..., :n_half, :n_half] -= softmax(zeta[..., None, :n_half], axis=-1)
        d_zeta_d_eta[..., n_half:, n_half:] -= softmax(zeta[..., None, n_half:], axis=-1)
        return Thresholds.d_tau(zeta) @ d_zeta_d_eta[..., 1:-1]
        # *** old method:
        # z_shape = (*eta.shape[:-1], 1)
        # v = np.concatenate((np.ones(z_shape), mapped_width(eta), np.ones(z_shape)),
        #                    axis=-1)
        # # = non-normalized widths
        # dv_deta = d_mapped_width_d_eta(v[..., 1:-1])  # variable intervals only
        # nv = v.shape[-1]
        # n_half = nv // 2
        # s1 = np.sum(v[..., :n_half], axis=-1, keepdims=True)
        # s2 = np.sum(v[..., n_half:], axis=-1, keepdims=True)
        # w = np.concatenate((v[..., :n_half] / s1, v[..., n_half:] / s2),
        #                    axis=-1)
        # # = normalized widths, separately in lower and upper intervals
        # dw_dv = np.zeros((*w.shape, v.shape[-1]))
        # dw_dv[..., range(n_half), range(n_half)] = 1. / s1[...]
        # dw_dv[..., range(n_half, nv), range(n_half, nv)] = 1. / s2[...]
        # dw_dv[..., :n_half, :n_half] -= v[..., :n_half, None] / s1[..., None, :]**2
        # dw_dv[..., n_half:, n_half:] -= v[..., n_half:, None] / s2[..., None, :]**2
        # dw_deta = dw_dv[..., :, 1:-1] * dv_deta[..., None, :]
        # cum_w = np.cumsum(w, axis=-1)
        # cw = cum_w[..., :-1, np.newaxis]  # only inner limits
        # sw = cum_w[..., -1:, np.newaxis]
        # # tau[..., m+1] = ln cw[..., m]  - ln (sw[..., 0] - cw[..., m])
        # # dcw_dw[..., m, i] = dcw[..., m] / dw[..., i]  = 1. if i <= m else 0.
        # nw = nv
        # dcw_dw = np.tril(np.ones((nw - 1, nw), dtype=int))
        # dtau_dw = dcw_dw / cw - (1 - dcw_dw) / (sw - cw)
        # dtau_d_eta = dtau_dw @ dw_deta
        # # dtau_d_eta.shape = (n_samples, nw - 1, n_eta); only finite thresholds
        # z_shape = (*dtau_d_eta.shape[:-2], 1, dtau_d_eta.shape[-1])
        # # include zero derivative for extreme -inf, +inf thresholds
        # d_tau_d_eta = np.concatenate((np.zeros(z_shape),
        #                        dtau_d_eta,
        #                        np.zeros(z_shape)), axis=-2)
        # # if not np.all(np.isclose(d_tau_new, d_tau_d_eta)):
        #     print('d_tau_new != d_tau_d_eta')
        # return d_tau_d_eta

    @staticmethod
    def tau_inv(tau):
        """Inverse of tau(eta)
        :param tau: 1D or 2D array with response thresholds, EXCEPT extremes at +-inf,
            i.e., all tau elements in (-inf, +inf),
            tau[..., m] = UPPER limit for m-th interval,
                = LOWER limit for the (m+1)-th interval
            tau.shape[-1] == M - 1 == number of response intervals - 1
        :return: eta: 1D or 2D array, such that
            self.tau(eta)[..., 1:-1] == tau, adjusted for mid tau value -> 0
        """
        n_categories = tau.shape[-1] + 1
        n_half = n_categories // 2
        # ensure tau[..., n_half - 1] -> 0
        tau = tau - tau[..., (n_half-1):n_half]
        # ensure local copy, with mid tau -> 0.
        eta = Thresholds.tau_inv(tau)
        eta[..., :n_half] -= eta[..., :1]
        eta[..., n_half:] -= eta[..., -1:]
        # with first and last elements == 0., excluded:
        return eta[..., 1:-1]


# ----------------------------- module help functions

# ----------- Original up to version 0.9.1 -> numeric overflow in some extreme cases
# mapped_width = _w = np.exp
# _w_inv = np.log  # _w_inv(w) -> eta, such that _w(eta) == w
#
# ---------- *** piecewise (exp, linear ) variant, tested no good
# ---------- *** piecewise (inverted linear, linear ) variant, tested no good


def mapped_width(eta):
    """Mapping function from model eta param to
    non-normalized interval widths in (0, 1) range
    :param eta: = 1D or 2D array with
        eta[..., m] = ...-th sample of parameter defining
            non-normalized width of m-th interval in mapped domain [0, 1].
        eta.shape[-1] == M == number of response-scale intervals.
    :return: w = array with mapped widths
        w.shape == eta.shape
    """
    return np.exp(eta) + ETA_W_EPSILON


def d_mapped_width_d_eta(w):
    """Derivative of mapped_width, as a function of w, NOT eta.
    :param w: = mapped_width(eta) = 1D or 2D array
    :return: array dw, with
        dw[..., m] = d _w(eta)[..., m] / d eta[..., m]
    """
    return w - ETA_W_EPSILON


def mapped_width_inv(w):
    """Inverse of mapped_width(eta),
    used only after each VI iteration,
    so it is OK to be slightly inconsistent with mapped_width()
    :param w: = 1D or 2D array with
        w[..., m] = ...-th sample of
            non-normalized width of m-th interval in mapped domain [0, 1].
        sum_m w[..., m] approx == 1.
        w.shape[-1] == M == number of response-scale intervals.
    :return: eta = array with mapped widths mapped back to eta parameter
        eta.shape == w.shape
    """
    # avoid log(0.) for numerical stability
    return np.log(w + np.finfo(float).tiny)


# ------------------------------------------------- TEST:
if __name__ == '__main__':
    # ******************** TEST extreme eta, too ********************
    from scipy.optimize import approx_fprime, check_grad
    n_samples = 3
    n_categories = 5
    print(f'n_categories = {n_categories}')

    test_w = np.ones(n_categories)
    test_w = 1. + np.arange(n_categories)

    test_w = np.tile(test_w, (n_samples, 1))
    print(f'test_w = {test_w}')

    for thr in [Thresholds, ThresholdsFree, ThresholdsMidFixed]:

        eta = np.log(test_w[..., :thr.n_param(test_w.shape[-1])])

        # test one extreme eta value:
        # eta[..., 0] = -50.

        print(f'\n*** Testing {thr.__name__}.tau ***')

        tau = thr.tau(eta)
        print(f'eta = ', eta)
        print(f'_w(eta) = ', mapped_width(eta))
        print(f'tau({eta}) = ', tau)
        print(f'tau_inv(tau[..., 1:-1] = ', thr.tau_inv(tau[..., 1:-1]))
        print(f'tau(tau_inv(tau[..., 1:-1]) = ', thr.tau(thr.tau_inv(tau[..., 1:-1])))

        # ----- **** check extreme tau, that might occur in restrict_xi ************
        print('*** Testing extreme tau:')
        tau = np.array([-np.inf, -100., 0., 0., 100., np.inf])
        print(f'tau = ', tau)
        print(f'tau_inv(tau[1:-1] = ', thr.tau_inv(tau[1:-1]))
        print(f'tau(tau_inv(tau[1:-1]) = ', thr.tau(thr.tau_inv(tau[1:-1])))

        def fun(eta):
            eta = np.tile(eta, (n_samples, 1))
            return thr.tau(eta)[0, limit]

        def jac(eta):
            eta = np.tile(eta, (n_samples, 1))
            return thr.d_tau(eta)[0, limit]

        if eta.ndim > 1:
            eta_test = eta[0]  # must be 1D vector for gradient test
        else:
            eta_test = eta

        for limit in range(1, n_categories):
            print(f'\n*** Testing Jacobian {thr.__name__}.d_tau[..., {limit}, :] ***')

            print(f'tau({eta_test}) = {thr.tau(eta_test)}')

            print('approx gradient = ', approx_fprime(eta_test, fun, epsilon=1e-3))
            print('exact  gradient = ', jac(eta_test))
            err = check_grad(fun, jac, eta_test, epsilon=1e-6)
            print('check_grad err = ', err)



"""This package implements a Bayesian probabilistic model for
analysis of Ecological Momentary Assessment (EMA) data.

EMA methodology is used, for example,
to evaluate the effect of any kind of psycho-social-medical intervention,
for example, the subjective performance of hearing aids or other equipment,
in the everyday life of the user or client.

For example, in an EMA study including hearing-aid users,
the participants might be asked to report, at random occasions in their daily life,
which type of activity they are currently engaged in,
or in which type of situation they are currently using their hearing aids.

The participants may also be asked to grade the performance of their current hearing aids,
for one or more perceptual Attributes, e.g., Speech Understanding and/or Sound Comfort.

Thus, EMA records usually include both NOMINAL and ORDINAL data.
Typically, many EMA records are collected from each participant,
but the number of records may vary a lot among respondents.

*** Template scripts:
run_ema.py
run_sim.py

*** Main Modules:
ema_data: defines classes EmaFrame, EmaDataSet, with read / write functions for EMA data,
    and some functions to show the distribution of raw data.

ema_model: defines class EmaModel as a probabilistic model for EMA data
    and a variational learning algorithm for all model parameters.

ema_group: defines class EmaGroupModel defining probabilistic model for
    ONE population and a group of EMA respondents recruited from that population.

ema_respondent: defines class EmaRespondentModel including recorded data and parameters
    for ONE EMA respondent.

ema_display: classes and functions to display analysis results

ema_display_format: functions formatting plots and tables with analysis results

ema_base: defines help class EmaParamBase defining internal properties and methods
    for accessing all model parameters.

*** Reference:
A Leijon, Petra von Gablenz, Inga Holube, Jalil Taghia and Karolina Smeds (2023):
Bayesian Analysis of Ecological Momentary Assessment (EMA) Data Collected in Adults
Before and After Hearing Rehabilitation.
Frontiers in Digital Health, 5(1100705). doi: 10.3389/fdgth.2023.1100705

*** Version History:
*Version 1.0.0:
2023-05-17, EmaModel.initialize(...) using n_participants_per_comp instead of max_n_comp.
2023-04-23, Respondent group specifications moved to ema_data.EmaFrame.
            EmaDataSet.load() signature changed: can find group category(-ies) in file path.
            EmaDataSet.save() signature changed: can save data in one single file,
                or in separate files, by group and / or participant

* Version 0.9.6:
Include range percentile plots and tables by groups, to show group differences

* Version 0.9.5:
Show observed and model-predicted attribute grade-counts,
    as requested by one reviewer of Frontiers (2023) paper.

Changed ema_base parameter extraction -> n_parameters == model degrees of freedom,
i.e., NO redundant model parameters.
This leads to a slight change of arbitrary scale zero point:
    If restrict_thresholds == True: one middle response threshold always forced -> zero,
    even for odd number of response categories.
    (Previous versions: mid-point of middle interval -> zero, if odd number of categories.)

Allow user to set number of samples for calculating population result displays.

* Version 0.9.4:
ema_display_format: Fixed filename bug that caused crash when running under Windows

* Version 0.9.3:
Ordinal rating scale may be tied to more than one Attribute.
changed names scenario -> situation,
'subject' -> 'participant',
'population_individual' -> 'random_individual'

* Version 0.9.2:
ema_data: improved optional overview of input data
ema_base: fixed numerical underflow problem in case of extreme response patterns.

* Version 0.9.1: Using Pandas DataFrame storage for input data, and all table results

* Version 0.8.3: changed ema_group classes to avoid unnecessary data transfer to Pool multi-processes

* Version 0.8.2: allow use multiprocessing.Pool for subjects in parallel

* Version 0.8.1: separate complete GMM for each group model,
    with mixture weights and components implemented in ema_group.EmaGroupModel

* Version 0.8: Improved variational approximation, less factorized,
    using joint q(xi_n, zeta_n) for EmaRespondentModel

* Version 0.7.2: Separate random Generators for each EmaRespondentModel

* Version 0.7.1: Minor updates:
    Allow seed for random generators, for reproducible numerical results
    Improved user control of simulated experiment, see script run_sim
    Minor change to ema_model.EmaRespondentModel.adapt_xi method

* Version 0.7:
2021-12-20, allow display of NAP effect measure of Attribute-Rating differences for
    individual participants, with point estimates and confidence intervals.
2021-12-17, display aggregated Attribute effects weighted by Scenario probabilities
2021-12-15, ema_model.EmaRespondentModel calculates individual Average Attribute Ratings, if desired.

* Version 0.6:
2021-12-08, allow user control of model restriction:
            restrict_attribute: sensory-variable location average forced -> 0.
            restrict_threshold: response-threshold median forced -> 0.

* Version 0.5.1: Minor update:
2021-11-25, allow EMA study with NO Attributes, i.e., ONLY Scenario probability profile(s)
2021-12-02, Minor fix: Attribute location == 0 given FIRST Scenario category
            regardless of regression_effect specification. (NO GOOD! Changed in v. 0.6)

* Version 0.5:
2021-10-12, Crude version, based on CountProfileCalc-2021, and PairedCompCalc (on PyPi),
2021-11-24, Functional beta version tested with simulated and (some) real data.
"""
__name__ = 'EmaCalc'
__version__ = '1.0.0'

__all__ = ['__version__', 'run_ema', 'run_sim']



"""This module defines a class for individual participant parameter model
to be part of an ema_group.EmaGroupModel instance,
which is part the main Bayesian probabilistic model of EMA data.

Individual parameter distributions are approximated by sampling.
The population mixture model is common prior for all individuals in
the participant group recruited from the same population.

*** Class Defined here:

EmaRespondentModel: Distribution of individual parameter vector
    assumed to determine the observed EMA data from ONE respondent,
    including, in each EMA record,
    (1) a nominal (possibly multidimensional) Situation category, and
    (2) ordinal Ratings for zero, one, or more perceptual Attributes.
    The parameter distribution is represented by an array xi with many samples.
    The sample distributions are independent across EmaRespondentModel instances,
    so instances may be adapt-ed in parallel processes.

*** Version History:
* Version 1.0.0:
2023-05-01, EmaRespondentModel.rvs_grade_count() reverted to old method from Frontiers paper.
            Alternative method from v.0.9.5 sometimes incorrect.
2023-04-26, corrected bug in EmaRespondentModel.rvs_grade_count(), in case multiple Phase-s
            Private property EmaRespondentModel._rng, using sampler rng, for reproducibility.

* Version 0.9.5:
2023-03-11, rvs_grade_count() uses model-PREDICTED situation count profiles,
            -> greater variability, compared to earlier OBSERVED counts for each situation
2023-03-03, new methods EmaRespondentModel.attribute_grade_count() and .rvs_grade_count()

* Version 0.9.3:
2022-07-27, changed module name ema_subject -> ema_respondent
            EmaSubjectModel -> EmaRespondentModel
2022-07-20, moved _initialize xi -> ema_base, for same reason
2022-07-19, moved logprob calc -> ema_base, to hide parameter indexing details there
2022-07-xx, minor update for notation change: scenario -> situation

* Version 0.9.2:
2022-06-15, EmaSubjectModel methods mean_attribute_grades, nap_diff deleted.
            Replaced by ema_data.EmaDataSet.(mean_attribute_table, nap_table).

2022-06-15, new _initialize_rating_eta(y); changed _initialize_rating_theta(y, eta)
            tested initial response thresholds crudely based on response counts

2022-05-21, EmaSubjectModel.cdf_arg: check for too small response interval,
            that might cause numerical underflow in case of many missing data

* Version 0.8.3:
2022-03-08, minor cleanup logging to work in multiprocessing

* Version 0.8.2: prepared for multi-processing subject adapt() in parallel processes
2022-03-03, EmaSubjectModel methods mean_zeta, mean_zeta_mom no longer needed

* Version 0.8.1: minor cleanup of comments and logger output

* Version 0.8
2022-02-12, Changed VI factorization for better approximation,
    with individual indicators conditional on parameter samples,
    defining variational q(zeta_n, xi_n) = q(zeta_n | xi_n) q(xi_n)
"""
import multiprocessing
import logging

import numpy as np
from scipy.optimize import minimize
import pandas as pd

from samppy import hamiltonian_sampler as ham
from samppy.sample_entropy import entropy_nn_approx as entropy

from EmaCalc.ema_base import PRIOR_PARAM_SCALE


# -------------------------------------------------------------------
__ModelVersion__ = "2023-05-01"

DITHER_PARAM_SCALE = 0.1 * PRIOR_PARAM_SCALE
# -> initial dithering of point-estimated individual parameters

N_SAMPLES = 1000
# = number of parameter vector samples in each EmaRespondentModel instance

logger = logging.getLogger(__name__)
# logger does NOT inherit parent handlers, when this module is running as child process
if multiprocessing.current_process().name != "MainProcess":
    # restore a formatter like ema_logging, only for console output
    console = logging.StreamHandler()
    console.setFormatter(logging.Formatter('{asctime} {name}: {message}',
                                           style='{',
                                           datefmt='%H:%M:%S'))
    logger.addHandler(console)
# Might instead send the parent logger as argument to adapt() ? ***
# logger.setLevel(logging.DEBUG)  # *** TEST
# ham.logger.setLevel(logging.DEBUG)  # *** TEST sampler


# -------------------------------------------------------------------
class EmaRespondentModel:
    """Container for EMA response-count data for ONE respondent,
    and a sampled approximation of the individual parameter distribution.

    Individual parameter distributions are approximated by a large set of samples
    stored as property xi, with
    self.xi[s, :] = s-th sample vector of parameters,
    with subsets of parameter types (alpha, beta, eta) as defined in self.base.
    """
    def __init__(self, base, situation_count, rating_count, xi, rng, prior=None, id=''):
        """
        :param base: single common ema_base.EmaParamBase object, used by all model parts
        :param situation_count: 2D array with response counts
            situation_count[k0, k] = number of responses
            in k-th <=> (k1, k2, ...)-th situation category at k0-th test phase,
            using flattened index for situation dimensions 1,2,....
            NOTE: ema_data.EmaFrame always stores test phase as first situation dimension.
        :param rating_count: list of 2D arrays with response counts
            rating_count[i][l, k] = number of responses for i-th ATTRIBUTE,
            at l-th ordinal level, given the k-th <=> (k0, k1, k2, ...)-th situation
        :param xi: 2D array with parameter sample vector(s)
            xi[s, j] = s-th sample of j-th individual parameter,
                concatenated by parameter sub-types as defined in base.
        :param rng: random Generator object for sampler
        :param prior: ref to owner EmaGroupModel
        :param id: key label of self
        """
        self.id = id
        self.base = base
        self.situation_count = situation_count
        self.rating_count = rating_count
        self.xi = xi
        self._sampler = ham.HamiltonianSampler(x=self.xi,
                                               fun=self._neg_ll,
                                               jac=self._grad_neg_ll,
                                               epsilon=0.2,
                                               n_leapfrog_steps=10,     # = default
                                               min_accept_rate=0.8,     # = default
                                               max_accept_rate=0.95,    # = default
                                               rng=rng
                                               )
        # keeping sampler properties across learning iterations
        self.prior = prior
        self.ll = None  # space for log-likelihood result from self.adapt()

    @property
    def _rng(self):
        return self._sampler._rng

    def __repr__(self):
        return (self.__class__.__name__ + '('
                + '\n\tsituation_count=' + f'{self.situation_count},'
                + '\n\trating_count=' + f'{self.rating_count},'
                + f'\n\txi= parameter array with shape {self.xi.shape}; id={id(self.xi)},'
                + f'\n\tid(prior)= {id(self.prior)},'
                + f'\n\tll= {self.ll})')

    @classmethod
    def initialize(cls, base, ema_df, rng, id=''):
        """Create model from recorded data
        :param base: single common EmaParamBase object
        :param ema_df: a pandas.DataFrame object for ONE respondent
        :param rng np.random.Generator for sampler use
        :param id: key to identify self (in logger output)
        :return: a cls instance
        """
        z = base.emf.count_situations(ema_df).reshape((base.emf.n_phases, -1))
        # z[k0, k] = number of EMA records at k0-th test phase
        # in k-th <=> (k1, k2, ...)-th situation, EXCL. k0= test phase
        y = [base.emf.count_grades(a, ema_df)
             for a in base.emf.attribute_dtypes.keys()]
        # y[i][l, k] = number of attribute_grades at l-th ordinal level for i-th ATTRIBUTE question
        # given k-th <=> (k0, k1, k2, ...)-th situation (INCL. k0= test phase)
        xi = base.initialize_xi(z, y)
        # dither to N_SAMPLES:
        xi = xi + DITHER_PARAM_SCALE * rng.standard_normal(size=(N_SAMPLES, len(xi)))
        return cls(base, z, y, xi, rng, id=id)

    def _kl_div_zeta(self, prior):
        """KLdiv for indicator variables zeta re prior
        = KLdiv{q(zeta | self.xi) || p(zeta | prior.mix_weight)}
        using current variational q(xi, zeta) = q(zeta | xi) q(xi)
        :param prior: ref to EmaGroupModel instance containing self
        :return: kl = scalar E{ ln q(zeta | self.xi) - ln p(zeta | prior.mix_weight)}

        Method: Leijon doc eq:VIlowerBoundCalc
        """
        w = prior.mean_conditional_zeta(self.xi)
        # w[c, s] = E{zeta_c | self.xi[s, :]} Leijon doc eq:ProbZetaGivenXi
        return np.sum(np.mean(w * (np.log(w + np.finfo('float').tiny)  # avoid log(0.) = nan
                                   - prior.mix_weight.mean_log[:, None]),
                              axis=1))

    def adapt(self, s_name):
        """Adapt parameter distribution self.xi
        to stored EMA count data, given the current self.w
        and the current estimate of population GMM components self.prior.comp.
        :param s_name: respondent id, for logger output  **** NOT NEEDED: using self.id ***
        :return: self, to send result via map() or Pool.imap()

        Result: Updated self.xi, and
            self.ll = E{ ln p(self.situation_count, self.rating_count | self.xi) }_q(xi)
                - E{ ln q(xi) / prior_p(xi) }_q(xi)
                - KLdiv( q(zeta | xi) || p(zeta | prior.mix_weight)
        """
        # find MAP point first:
        # lp_0 = - np.mean(self._neg_ll(self.xi), axis=0)  # ****** ref for test
        xi_0 = np.mean(self.xi, axis=0)
        res = minimize(fun=self._neg_ll,
                       jac=self._grad_neg_ll,
                       x0=xi_0)
        if res.success:
            xi_map = res.x.reshape((1, -1))
        else:
            raise RuntimeError(f'{self.id}: MAP search did not converge: '
                               + 'res= ' + repr(res))
        if len(self.xi) != N_SAMPLES:
            # run sampler starting from x_map
            self._sampler.x = xi_map
        else:
            # we have sampled before, start from those samples
            self._sampler.x = self.xi + xi_map - xi_0
        # **** ------------------------------------- test effect of MAP
        # lp_1 = - np.mean(self._neg_ll(self._sampler.x), axis=0)  # ***** test after MAP adjustment
        # print(f'adapt: Subj {s_name}: MAP adjustment: d_lp = {lp_1 - lp_0:.3f}')
        # -----------------------------------------------------
        self._sampler.safe_sample(n_samples=N_SAMPLES, min_steps=2)
        logger.debug(f'{self.id}: sampler.epsilon= {self._sampler.epsilon:.3f}. '
                     + f'accept_rate= {self._sampler.accept_rate:.1%}. '
                     + f'n_steps= {self._sampler.n_steps}')
        self.xi = self._sampler.x
        self.base.restrict(self.xi)
        # adjust for modified xi:
        self._sampler.U = self._sampler.potential(self._sampler.x)
        self._sampler.args = ()  # just in case a pickled prior copy was there
        # Calc log-likelihood contribution with final xi samples:
        lp_xi = - np.mean(self._sampler.U)  # after base.restrict
        # lp_xi = E_xi{ ln p(data | xi) + self.prior.logpdf(xi) }
        h_xi = entropy(self.xi)
        # approx = - E{ ln q(xi) }
        kl_zeta = self._kl_div_zeta(self.prior)
        self.ll = lp_xi + h_xi - kl_zeta
        logger.debug(f'{self.id}: adapt: ll={self.ll:.3f}; '
                     + f'(lp_xi={lp_xi:.3f}; '
                     + f'h_xi= {h_xi:.3f}. '
                     + f'-kl_zeta= {-kl_zeta:.3f})')
        self.ll = lp_xi + h_xi - kl_zeta
        return self

    def _neg_ll(self, xi):
        """Objective function for self.adapt_xi
        :param xi: 1D or 2D array of candidate parameter vectors
        :return: neg_ll = scalar or 1D array
            neg_ll[...] = - ln P{ self.situation_count, self.rating_count | xi[..., :]}
                    - ln p(xi[..., :] | prior)
            neg_ll.shape == xi.shape[:-1]
        """
        return - self.prior.logpdf(xi) - self.logprob(xi)

    def _grad_neg_ll(self, xi):
        """Gradient of self._neg_ll w.r.t. xi
        :param xi: 1D or 2D array of candidate parameter vectors
        :return: dll = scalar or 1D array
            dll[..., j] = d _neg_ll(xi[..., :]) / d xi[..., j]
            dll.shape == xi.shape
        """
        return - self.prior.d_logpdf(xi) - self.d_logprob(xi)

    def logprob(self, xi):
        """log likelihood of EMA count data, given parameters xi
        :param xi: 1D or 2D array of candidate parameter vectors
        :return: ll = scalar or 1D array
            ll[...] = ln P{ self.situation_count, self.rating_count | xi[..., :]}
            ll.shape == xi.shape[:-1]
        """
        ll_z = self.base.situation_logprob(xi, self.situation_count)
        ll_y = self.base.rating_logprob(xi, self.rating_count)

        # ----------- any rating_logprob = -inf -> ll_y == NaN
        if np.any(np.isnan(ll_y)):
            logger.warning('EmaRespondentModel.logprob: Some ll_y == NaN. Should never happen!')
        return ll_z + ll_y

    def d_logprob(self, xi):
        """Gradient of self.logprob(xi)
        :param xi: 1D or 2D array with candidate parameter vectors
        :return: d_ll = 1D or 2D array
            d_ll[..., j] = d ln P{ self.situation_count, self.attribute_grades | xi[..., :]} / d xi[..., j]
            d_ll.shape == xi.shape
        """
        d_ll_z = self.base.d_situation_logprob(xi, self.situation_count)
        # = list of arrays with shapes concatenable along axis=-1
        d_ll_y = self.base.d_rating_logprob(xi, self.rating_count)
        # = list of arrays with shapes concatenable along axis=-1
        return np.concatenate(d_ll_z + d_ll_y, axis=-1)

    def rvs(self, size=N_SAMPLES):
        # re-sample if size != len(self.xi) ???
        return self.xi

# ------------------------------------- display functions
    def attribute_grade_count(self, a, groupby=None):
        """Collect table of ordinal grade counts for ONE attribute,
        for this participant, optionally subdivided by situation.
        Similar to ema_data.EmaDataSet.attribute_grade_count()
        :param a: selected attribute key
        :param groupby: (optional) single situation dimension or list of such keys
            for which separate attribute-counts are calculated.
            Counts are summed across any OTHER situation dimensions.
        :return: if groupby is defined:
            a pd.DataFrame object with all grade counts,
            with one row for each (*groupby) combination
            and one column for each grade category
            if groupby is empty: a pd.Series with sum of all grade counts
        """
        emf = self.base.emf
        a_ind = list(emf.attribute_scales.keys()).index(a)
        sit_index = pd.MultiIndex.from_product([sit_dtype.categories
                                                for sit_dtype in emf.situation_dtypes.values()],
                                               names=emf.situation_dtypes.keys())
        a_grades = emf.ordinal_scales[emf.attribute_scales[a]].categories
        df = pd.DataFrame(self.rating_count[a_ind].T,
                          index=sit_index, columns=a_grades)
        df.columns.set_names(a, inplace=True)
        if groupby is None:
            return df.sum()
        else:
            return df.groupby(level=groupby, sort=False, group_keys=True).sum()

    def rvs_grade_count(self, a, groupby=None, sample_head='_Sample'):
        """Calculate sampled model-predicted distribution of response counts
        for comparison with observed count histograms.
        :param a: key for selected attribute
        :param groupby: (optional) single situation key (dimension) or list of such keys
            for which separate attribute-counts are calculated.
            Counts are summed across any OTHER situation dimensions.
        :param sample_head: (optional) header name for sample index
        :return: a pd.DataFrame object with all grade counts,
            with one row for each (sample, *groupby) multi-index combination
            and one column for each grade category.
        """
        # *** use total observed count to estimate situation-count distribution !
        emf = self.base.emf
        a_index = list(emf.attribute_dtypes.keys()).index(a)
        if groupby is None:
            groupby = []
        elif isinstance(groupby, str):
            groupby = [groupby]
        theta = self.base.attribute_theta(self.xi, a)
        # theta[s, k0, k1,...] = s-th sample of attribute a, given (k0, k1,...)-th situation
        theta_shape = theta.shape
        theta = theta.reshape((theta_shape[0], -1))
        # theta[s, k] = s-th sample of attribute a, given k-th <=> (k0, k1,...)-th situation
        tau = self.base.attribute_tau(self.xi, a)
        tau = np.concatenate((np.full((tau.shape[0], 1), -np.inf),
                              tau,
                              np.full((tau.shape[0], 1), np.inf)), axis=-1)
        # tau[s, l] = s-th sample of l-th median rating threshold for attribute a
        arg_low = tau[..., :-1, None] - theta[..., None, :]
        # a[s, l, k] = lower interval limits for l-th category in k-th situation
        arg_high = tau[..., 1:, None] - theta[..., None, :]  # upper interval limits
        lp = self.base.rv_class.log_cdf_diff(arg_low, arg_high)
        p_response = np.exp(lp).transpose((0, 2, 1))
        # p[s, k, l] = s-th sample of prob of l-th response in k-th situation
        # p_sit = self.base.situation_prob(self.xi)
        # # p_sit[s, k0, k1, ...] = s-th sample of CONDITIONAL prob for (k1,...)-th situation, given k0 (Phase)
        # p_sit /= p_sit.shape[1]  # equal weight for all Phase-s
        # # p_sit[s, k0, k1,...] = s-th sample of prob for k-th <-> (k0, k1,...)-th situation
        # p_sit = p_sit.reshape((len(p_sit), -1))
        # # p_sit[s, k] = s-th sample of prob for k-th situation, across ALL attributes
        # c_tot = np.sum(self.rating_count[a_index])
        # # c_tot = total count across all situations
        # c_sit_all = self._rng.multinomial(c_tot, p_sit)
        # c_sit_all[s, k] = s-th sample of PREDICTED count in k-th situation across ALL attributes
        # *** This gives WRONG result in case the attribute is relevant only in some situation,
        # *** E.g., in the IHAB study, attribute Involvement has responses only in CoSS=C2.
        c_sit = np.sum(self.rating_count[a_index], axis=0)
        # c_sit[k] = total OBSERVED count of responses in k-th situation for THIS attribute!
        # *** This is the method used in the Frontiers paper!
        n_count = self._rng.multinomial(c_sit, pvals=p_response)
        # n_count[s, k, l] = s-th sample of counts in l-th grade category in k-th situation
        df_index = pd.MultiIndex.from_product([range(len(n_count)),
                                               *[sit_dtype.categories
                                                 for sit_dtype in emf.situation_dtypes.values()]],
                                              names=[sample_head, *emf.situation_dtypes.keys()])
        a_grades = emf.ordinal_scales[emf.attribute_scales[a]].categories
        df = pd.DataFrame(n_count.reshape((-1, n_count.shape[-1])),
                          index=df_index, columns=a_grades)
        df.columns.set_names(a, inplace=True)
        return df.groupby(level=[0, *groupby], sort=False, group_keys=True).sum()



"""General utility classes and functions for internal EmaCalc use.

Defines class
EmaParamBase --- container for common properties used by all model classes
    specifying the indexing of separate types of model parameters,
    calculating logprob for observed data, given tentative model parameters.

NOTE: To make the model identifiable, some restriction is necessary.
The zero point on the attribute scale is otherwise arbitrary,
if both latent-variable locations and response thresholds are freely variable.

This behavior is user-controlled by initialization arguments
restrict_attribute and restrict_threshold.

The model is also slightly restricted by a weakly informative prior,
for numerical stability in case of extreme response patterns,
e.g., aLL ratings in the highest ordinal category.


*** Version history:
* Version 0.9.6:
2023-04-02, new access functions situation_prob_df, attribute_theta_df yielding pandas result

* Version 0.9.5:
2022-11-28, changed _theta_map() to define attribute theta directly from effect parameters
            depending on restrict_attribute setting.
            One beta parameter less, if restrict_attribute is True.
            EmaParamBase._restrict_attr() no longer needed.
2022-11-19, threshold calculations done by class methods in module ema_thresholds

* Version 0.9.3:
2022-08-12, allowing ordinal scale tied to more than one attribute
2022-08-02, changed EmaParamBase properties specifying index slices
2022-07-21, xi parameters stored in order (alpha, beta, eta) parameter groups
2022-07-15, separate beta_slices, eta_slices;
            to allow threshold parameters tied to more than one attribute
2022-07-13, changed all names involving scenario -> situation

* Version 0.9.2:
2022-06-12, response-threshold mapping functions
    mapped_width(), d_mapped_width_d_eta(), and mapped_width_inv()
    slightly modified to be safe against numerical underflow.
    Global PRIOR_PSEUDO_RESPONDENT slightly modified, with no apparent effect on result.

* Version 0.9:
2022-03-17, use Pandas CategoricalDtype instances in EmaFrame situation_dtypes and attributes
2022-03-18, use Pandas DataFrame for EMA data in EmaDataSet, to allow many input file formats

* Version 0.8.1
2022-02-28, changed class name -> EmaParamBase
2022-02-23, GMM comp moved to ema_model.EmaGroupModel, separate GMM for each group

* Version 0.7.1
2022-01-11, _make_prior with log normalized uniform scenario prob

* Version 0.7
2022-01-02, minor cleanup _make_prior: ONLY scalar precision shape for GMM components

* Version 0.6
2021-12-03, user-definable prior inter-individual variance for reference parameters
2021-12-05, testing ways to set hyper-prior
2021-12-06, boolean properties restrict_attribute, restrict_threshold of EmaParamBase
            for user-specified model restriction
2021-12-08, minor checks against numerical overflow

* Version 0.5.1
2021-11-26, allow ZERO Attributes, i.e., empty emf.attribute_grades
2021-12-01, cleanup some doc comments
2021-12-02, Attribute sensory location fixed == 0. for first Scenario category
            regardless of regression_effect specification.

* Version 0.5
2021-11-24, first published beta version
"""
# *** Use only K - 1 alpha params for each phase, no redundant parameters?
# *** e.g., with only alpha differences stored in xi ?
import numpy as np
from itertools import chain
from collections import namedtuple
from scipy.special import logit, expit
from scipy.special import logsumexp, softmax
import logging
import pandas as pd

from EmaCalc.gauss_gamma import GaussianRV
# from EmaCalc.ema_thresholds import ThresholdsOld  # for backward compatibility
from EmaCalc.ema_thresholds import ThresholdsFree, ThresholdsMidFixed

# from EmaCalc.ema_display import aggregate_situation_prob  # ********* for TEST

# ------------------------------------------------------
__version__ = "2023-05-13"

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)  # *** TEST

PRIOR_PSEUDO_RESPONDENT = 0.5  # seems to work OK
# = hyperprior total pseudo-count re ONE real respondent
# = prior GaussianRV.mean.learned_weight for all GMM components.
# This value is NOT critical for Occam's Razor behavior.

PRIOR_PARAM_SCALE = 1.
# = main hyperprior scale of most Gaussian model parameters,
# defined in Thurstone d-prime (probit) units for attribute parameters,
# rescaled if the Bradley (logistic) model is used for latent variables.
# This prior may have effect on the Occam's Razor function:
# small values may allow several active GMM components with small variance,
# but experiments suggest the prior value is not critical.

# *** 2023-05-16, tested with PRIOR_PARAM_SCALE = 0.3:
# *** Too small, -> some GMM population-model comp. with only ONE member.

PRIOR_PREC_A = PRIOR_PSEUDO_RESPONDENT / 2
# = GaussianRV precision shape for ALL parameters

# PRIOR_PREC_A_THRESHOLDS = 4 * PRIOR_PREC_A
# *** Allow user to set smaller prior variance for threshold parameters ?
# *** 2023-05-13 TESTED, does not make clear difference to GMM number of components.

PRIOR_PREC_B = PRIOR_PARAM_SCALE**2 / 2
# = GaussianRV precision inv_scale for MOST EmaModel parameters
# -> allows very small precision <=> large inter-individual variance
# -> mode of prior component-element variance = PRIOR_PREC_B /(PRIOR_PREC_A + 1) approx= PRIOR_PREC_B

ETA_W_EPSILON = np.finfo(float).eps
# = additive constant to prevent too small mapped_width(eta)

# ------------------------------------------------------- Help classes
AttributeSlice = namedtuple('AttributeSlice', ['beta_slice', 'eta_slice'])


class Slicer:
    """Help class to create consecutive slices for parts of parameter vector
    """
    def __init__(self, start=0):
        self.start = start

    def slice(self, length):
        """Create a new slice object tight after previous slice
        :param length: desired slice size
        :return: a slice object
        """
        stop = self.start + length
        sl = slice(self.start, stop)
        self.start = stop  # for next call
        return sl


# ------------------------------------------------------------------
class EmaParamBase:
    """Container for common properties of ema_model.EmaModel and its sub-models.

    Each individual parameter distribution is represented by
    a large set of samples, stored as property EmaRespondentModel.xi, with
    xi[s, :] = s-th sample of the parameter vector for ONE respondent.

    All model classes share mapping properties defined here,
    for extracting parameter subsets of three different types
    from a xi array of parameter vectors.
    Parameter types are stored consecutively in order (alpha, beta, eta)
    """
    def __init__(self, emf, effects, rv_class,
                 thr,
                 theta_map,
                 alpha_slices,
                 attribute_slices,
                 tied_scales,
                 comp_prior,
                 restrict_attribute=False,
                 restrict_threshold=False):
        """
        :param emf: single ema_data.EmaFrame instance,
        :param effects: iterable with regression-effect terms for attribute regression
            effects[i] = single key in emf.situation_dtypes.keys() or tuple of such keys
        :param rv_class: latent sensory variable class, defined in module ema_latent,
            defining the distribution as either logistic or normal (Gaussian).
        :param thr: class defining threshold calculations, defined in module ema_thresholds,
        :param theta_map: fixed 2D array to extract latent-variable location samples from xi array
        :param alpha_slices: list of slice objects, such that
            xi[:, alpha_slices[t]] = alpha = log-likelihood for situations in t-th test phase
        :param attribute_slices: list of AttributeSlice objects, ordered like emf.attribute_dtypes,
            with attribute_slices[i] = (beta_slice, eta_slice), such that
            xi[:, beta_slice] = beta = regression-effect parameters for i-th Attribute
            xi[:, eta_slice] = eta = threshold parameters for i-th Attribute
        :param tied_scales: list of indices into attribute_slices,
            to allow more than one attribute to use SAME ordinal scale
        :param comp_prior: single gauss_gamma.GaussianRV instance,
            prior for ALL GMM components in ema_group.PopulationModel instance
        :param restrict_attribute: (optional) boolean switch
            to force restriction on attribute sensory-variable location mean across situations
        :param restrict_threshold: (optional) boolean switch
            to force restriction on middle response-threshold location
        """
        self.emf = emf
        self.effects = effects
        self.rv_class = rv_class
        self.thr = thr
        self.theta_map = theta_map
        self.alpha_slices = alpha_slices
        self.attribute_slices = attribute_slices
        self.tied_scales = tied_scales
        self.comp_prior = comp_prior
        self.restrict_attribute = restrict_attribute  # needed? already defined in theta_map
        self.restrict_threshold = restrict_threshold  # needed? already defined by class thr

    def __repr__(self):
        return (self.__class__.__name__ + '(' +
                '\n\t'.join(f'{k}= {repr(v)},'
                            for (k, v) in self.__dict__.items()) +
                '\n\t)')

    @classmethod
    def initialize(cls, emf, effects, rv_class,
                   restrict_attribute=False,
                   restrict_threshold=False):
        """Assign all parameter-extraction properties, and
        prior mixture components as GaussianRV objects with correct size.
        :param emf: single ema_data.EmaFrame instance,
            defining the experiment structure.
        :param effects: iterable with regression-effect terms for attribute regression
            effects[i] = single key in emf.situation_dtypes.keys() or tuple of such keys
        :param rv_class: latent variable class, defined in ema_latent,
            defining its distribution as either logistic or normal.
        :param restrict_attribute: (optional) boolean switch
            to force restriction on attribute sensory-variable locations
        :param restrict_threshold: (optional) boolean switch
            to force restriction on response-threshold locations
        :return: a cls instance

        Result: all properties initialized
        NOTE: xi parameter slices in order: all alpha, all beta, all eta
        """
        _check_effects(emf, effects)
        effects = [e_i if type(e_i) is tuple else (e_i,)
                   for e_i in effects]
        # = list of requested regression effects of situation categories
        # to be estimated for their influence on Attribute responses.
        # Each element MUST be a tuple with one or more situation key(s).

        # Define mapping indices from rows of xi array to its parts:
        n_phases = emf.situation_shape[0]
        n_situations = np.prod(emf.situation_shape[1:], dtype=int)  # within each phase
        slice_gen = Slicer()  # creating adjacent slice objects
        alpha_slices = [slice_gen.slice(n_situations)
                        for _ in range(n_phases)]
        # = list of slice objects, such that
        # xi[:, alpha_slices[t]] = alpha = log-likelihood for situations in t-th test phase
        # i.e., n_phases slices with equal lengths, for all test phases.

        theta_map = _theta_map(emf, effects, restrict_attribute)
        # = fixed 2D array to extract latent-variable location samples from xi array
        # by method attribute_theta(...)

        n_beta = theta_map.shape[0]
        # = number of regression-effect parameters, same for every Attribute
        n_attributes = len(emf.attribute_dtypes)
        beta_slices = [slice_gen.slice(n_beta)
                       for _ in range(n_attributes)]
        # = list of slice objects, such that
        # xi[:, beta_slices[i]] = beta = effect parameters for i-th attribute, such that
        # latent-variable location theta = np.dot(beta, theta_map)

        if restrict_threshold:
            thr = ThresholdsMidFixed
        else:
            thr = ThresholdsFree
        # Number of eta parameters depends on class thr
        unique_slices = {scale_id: slice_gen.slice(thr.n_param(len(scale.categories)))
                         for (scale_id, scale) in emf.ordinal_scales.items()}

        eta_slices = [unique_slices[scale_id]
                      for scale_id in emf.attribute_scales.values()]
        # = list of slice objects, such that
        # xi[:, eta_slices[i]] = eta = threshold parameters for i-th attribute, such that
        # response thresholds tau = thr.tau(eta)
        # Some thresholds may be identical for more than one attribute
        attribute_slices = [AttributeSlice(beta, eta)
                            for (beta, eta) in zip(beta_slices,
                                                   eta_slices)]
        # *** attribute_slices as dict instead ?
        tied_scales = None
        if len(attribute_slices) > len(unique_slices):
            attribute_keys = list(emf.attribute_scales.keys())
            tied_scales = [[attribute_keys.index(a_key) for a_key in a_keys]
                           for a_keys in emf.scale_attributes.values()]
        # tied_scales[s] = list of attribute indices a_i, such that
        # attribute_slices[a_i].eta_slice is eta_slices[s],
        # allowing same eta parameters to be used for more than one attribute

        n_parameters = slice_gen.start  # no more slices needed
        comp_prior = _make_prior(n_parameters,
                                 alpha_slices,
                                 attribute_slices,
                                 rv_class)
        # = single GaussianRV instance, prior for ALL GMM components,
        # fixed throughout the VI learning procedure.
        logger.debug(f'PRIOR_PSEUDO_RESPONDENT = {PRIOR_PSEUDO_RESPONDENT:.3f}; ' +
                     f'PRIOR_PARAM_SCALE = {PRIOR_PARAM_SCALE:.3f}; ' +
                     f'PRIOR_PREC_A = {PRIOR_PREC_A:.3f}; ' +
                     f'PRIOR_PREC_B = {PRIOR_PREC_B:.3f}')
        logger.debug(f'comp_prior.prec.a = '
                     + np.array_str(comp_prior.prec.a,
                                    precision=3))
        logger.debug('comp_prior.prec.b = '
                     + np.array_str(comp_prior.prec.b,
                                    precision=5))
        logger.debug('mode{1 / comp_prior.prec}= '
                     + np.array_str(comp_prior.prec.mode_inv(),
                                    precision=5))
        return cls(emf, effects, rv_class,
                   thr,
                   theta_map,
                   alpha_slices,
                   attribute_slices,
                   tied_scales,
                   comp_prior,
                   restrict_attribute,
                   restrict_threshold)

    @property
    def n_parameters(self):  # *** needed ?
        if len(self.attribute_slices) > 0:
            return max(a_slice.eta_slice.stop
                       for a_slice in self.attribute_slices)
        else:
            return self.alpha_slices[-1].stop

    # def beta_0_size(self):  # needed only for old _restrict_attr
    #     """number of effect parameters for FIRST effect term
    #     :return: n_beta = scalar
    #
    #     NOTE: remaining effect terms are relative to first effect.
    #     """
    #     return np.prod([len(self.emf.situation_dtypes[sc_i].categories)
    #                     for sc_i in self.effects[0]],
    #                    dtype=int)

    def situation_prob(self, xi):
        """Extract probability-mass for situations, given parameters,
        used mainly by ema_display
        :param xi: 2D array with parameter samples
        :return: u = mD array with situation probability-mass within each phase,
            u[s, k0, k1, k2, ...] = s-th sample of P[(k1, k2,...)-th situation | phase k0]
            sum u[s, k0] == 1., for all s and k0

        2022-02-15, check for underflow
        """
        n_sc = self.alpha_slices[-1].stop
        alpha = xi[:, :n_sc].reshape((xi.shape[0], self.emf.n_phases, -1))
        alpha -= np.amax(alpha, axis=-1, keepdims=True)
        too_small = np.all(alpha < np.log(np.finfo(float).tiny), axis=-1)
        n_underflow = np.sum(too_small)
        if n_underflow > 0:
            logger.warning(f'situation_prob: {n_underflow} alpha samples too small. '
                           + 'Should not happen. Maybe too few responses?')
            logger.debug(f'alpha[too_small, :] = {alpha[too_small]}')
        u = np.exp(alpha)
        # avoid nan after normalization, should not be needed !
        u /= np.sum(u, axis=-1, keepdims=True)  # normalize within each Phase
        return u.reshape((-1, *self.emf.situation_shape))

    def situation_prob_df(self, xi, groupby=None, sample_head='_sample'):
        """Extract probability-mass for situations, given parameters,
        used mainly by ema_display
        :param xi: 2D array with parameter samples
        :param groupby: (optional) tuple with situation key(s) to be included.
            If undefined, include ALL situation dimensions as defined in self.emf
        :param sample_head: (optional) level name of sample index
        :return: ds = pandas Series object
            with a MultiIndex with levels (sample_head, *groupby),
            containing the CONDITIONAL probability for categories in groupby[0], given other groupby cases,
            aggregated across situation dimensions NOT included in groupby.
        """
        u = self.situation_prob(xi)
        # u[s, k0, k1, k2, ...] = s-th sample of conditional P[(k1, k2,...)-th situation | phase k0]
        df_index = pd.MultiIndex.from_product([range(len(u)),
                                               *[sit_dtype.categories
                                                 for sit_dtype in self.emf.situation_dtypes.values()]],
                                              names=[sample_head, *self.emf.situation_dtypes.keys()])
        if groupby is None:
            return pd.Series(u.reshape((-1,)), index=df_index)
        else:
            u /= u.shape[1]
            # u[s, k0, k1, k2, ...] = s-th sample of prob-mass P[(k0, k1, k2,...)-th situation]
            df = pd.Series(u.reshape((-1,)), index=df_index)
            df = df.groupby(level=[0, *groupby], sort=False, group_keys=True).sum()
            # **** needed only if len(groupby) > 1
            if len(groupby) > 1:
                # scale to CONDITIONAL prob for categories in groupby[0], given other dimensions
                df = df / df.groupby(level=[0, *groupby[1:]],
                                     sort=False, group_keys=True).transform(sum)
                # CHECK: df.groupby(level=[0, *groupby[1:]], sort=False).sum() == 1., in all groups
            return df

    def attribute_theta(self, xi, a):
        """Extract location of latent sensory variable, for ONE given attribute
        used only by ema_display
        :param xi: 2D array with parameter sample vectors
            xi[s, :] = s-th parameter sample vector
        :param a: attribute key = one of self.emf.attribute_grades.keys()
        :return: theta = mD array, with
            theta[s, k0, k1, ...] = s-th sample of
                attribute location, given the (k0, k1, ...)-th situation.
        """
        a_index = list(self.emf.attribute_dtypes.keys()).index(a)
        beta_slice = self.attribute_slices[a_index].beta_slice
        # *** attribute_slices as dict instead ?
        beta = xi[..., beta_slice]
        return np.dot(beta, self.theta_map).reshape((-1, *self.emf.situation_shape))

    def attribute_theta_df(self, xi, a, groupby=None, sample_head='_sample'):
        """Extract location of latent sensory variable, for ONE given attribute,
        given each situation category in selected situation dimension(s).
        Used mainly by ema_display
        :param xi: 2D array with parameter sample vectors
            xi[s, :] = s-th parameter sample vector
        :param a: attribute key = one of self.emf.attribute_grades.keys()
        :param groupby: (optional) tuple with situation key(s) to be included.
            If undefined, include ALL situation dimensions as defined in self.emf
        :param sample_head: (optional) level name of sample index
        :return: ds = pandas Series object
            with a MultiIndex with levels (sample_head, *groupby),
            containing the Attribute value samples for each category in groupby situation dimension(s),
            aggregated across all OTHER situation_dtypes not included in groupby,
            weighted by AVERAGE situation probabilities in those dimensions.
        """
        theta = self.attribute_theta(xi, a)
        # theta[s, k0, k1, ...] = s-th sample of attribute location, in (k0, k1, ...)-th situation.
        df_index = pd.MultiIndex.from_product([range(len(theta)),
                                               *[sit_dtype.categories
                                                 for sit_dtype in self.emf.situation_dtypes.values()]],
                                              names=[sample_head, *self.emf.situation_dtypes.keys()])
        theta_ds = pd.Series(theta.reshape((-1,)), index=df_index)
        if groupby is None:
            return theta_ds
        else: # weighted average across situation dimensions not included
            aggregate_by = list(set(self.emf.situation_dtypes.keys()) - set(groupby))
            # = OTHER situation dimensions to be aggregated out
            u = self.situation_prob_df(xi, sample_head=sample_head)  # *** use array access ??? ***
            # u = Series object with same MultiIndex as theta_ds),
            # containing the CONDITIONAL probability for categories in situation_dtypes[1:]
            # given Phase category in situation_dtypes[0]
            u = u / len(u.index.levels[1])
            # u = absolute prob.mass for each sample and situation category
            # w_cond = u / u.groupby(level=[0, *groupby], sort=False).transform(sum)
            # # = CONDITIONAL prob.mass for categories in aggregate_by, GIVEN each category in groupby
            w_av = u.groupby(level=[0, *aggregate_by],
                             sort=False, group_keys=True).transform(sum)
            # = AVERAGE prob.mass for categories in aggregate_levels
            # = same as ema_display.aggregate_situation_theta
            # -> slightly less variability than w_cond
            theta_ds = theta_ds * w_av  # w_cond  # ???
            return theta_ds.groupby(level=[0, *groupby], sort=False, group_keys=True).sum()

    def attribute_tau(self, xi, a):
        """Extract response thresholds for ONE given attribute,
        EXCEPT fixed extreme limits at -inf, +inf.
        Used only by ema_display
        :param xi: 2D array with parameter sample vectors
            xi[s, :] = s-th parameter sample vector
        :param a: attribute key = one of self.emf.attribute_grades.keys()
        :return: tau = 2D array, with
            tau[s, l] = s-th sample of UPPER limit of l-th response interval,
                EXCEPT the last at +inf
                tau.shape[-1] == len(self.emf.attribute_grades[a]) - 1
        """
        a_index = list(self.emf.attribute_dtypes.keys()).index(a)
        eta_slice = self.attribute_slices[a_index].eta_slice
        eta = xi[..., eta_slice]
        return self.thr.tau(eta)[..., 1:-1]

    # ---------------------- methods for logprob calculation:

    def situation_logprob(self, xi, situation_count):
        """Situation logprob given tentative parameter array
        Used by ema_respondent.EmaRespondentModel.adapt().
        :param xi: 1D or 2D array of tentative parameter samples
        :param situation_count: 2D array with response counts,
            situation_count[k0, k] = number of responses
            in k-th <=> (k1, k2, ...)-th situation category at k0-th test phase,
            using flattened index for situation dimensions 1,2,....
        :return: ll = scalar or 1D array
            ll[...] = ln P{situation_count | xi[..., :]}
            ll.shape == xi.shape[:-1]
        """
        def sit_lp(alpha):
            """
            :param alpha: non-normalized log-prob-mass for ONE test phase
            :return: lp = 1D or 2D array
                lp[..., k] = log P{visit k-th situation | alpha[..., :] }
                normalized as sum_k exp(lp[..., k]) == 1.
            """
            return alpha - logsumexp(alpha, axis=-1, keepdims=True)
        # -------------------------------------------------------
        return sum(np.dot(sit_lp(xi[..., alpha_ind]), z_t)
                   for (z_t, alpha_ind) in zip(situation_count,
                                               self.alpha_slices))

    def d_situation_logprob(self, xi, situation_count):
        """Gradient of situation_logprob,
        to be concatenated with d_rating_logprob(...)
        :param xi: 1D or 2D array of tentative parameter samples
        :param situation_count: 2D array with response counts
            situation_count[k0, k] = number of responses
            in k-th <=> (k1, k2, ...)-th situation category at k0-th test phase,
            using flattened index for situation dimensions 1,2,....
        :return: list of 1D or 2D arrays ll_t, with
            ll_t[..., j] = d ln P{situation_count[t] | xi[..., :]} / d xi[..., j_t]
                for j_t-th alpha parameter in t-th test phase
            ll_t.shape[-1] == situation_count.shape[-1]
        """
        def d_sit_lp(alpha):
            """Gradient of sit_lp(alpha)
           :param alpha: non-normalized log-prob-mass for ONE test phase
            :return: d_lp = 1D or 2D array with
                d_lp[..., k] = log P{visit k-th situation | alpha[..., :] }
                d_lp.shape == alpha.shape
            """
            return np.eye(alpha.shape[-1]) - softmax(alpha[..., None, :], axis=-1)
            # -------------------------------------------------------------------

        # return [np.einsum('k, ...kj -> ...j',  # ******* use dot or @?
        #                   z_t, d_sit_lp(xi[..., alpha_ind]))
        return [np.dot(z_t, d_sit_lp(xi[..., alpha_ind]))
                for (z_t, alpha_ind) in zip(situation_count,
                                            self.alpha_slices)]

    def rating_logprob(self, xi, rating_count):
        """Rating logprob given tentative parameter array
        Used by ema_respondent.EmaRespondentModel.adapt().
        :param xi: 1D or 2D array of tentative parameter samples
        :param rating_count: list of 2D arrays with response counts
            rating_count[i][l, k] = number of responses for i-th ATTRIBUTE,
            at l-th ordinal level, given the k-th <=> (k0, k1, k2, ...)-th situation
        :return: ll = scalar or 1D array
            ll[...] = log P{rating_count | xi[..., :]}
            ll.shape == xi.shape[:-1]
        """
        def rating_lp(beta, eta):
            """log-prob for ONE attribute
            :param beta: 1D or 2D array with effect params
            :param eta: 1D or 2D array with threshold params
            :return: lp = 2D or 3D array with
                lp[..., l, k] = ln P{l-th grade | k-th situation, beta, eta}
            """
            return self.rv_class.log_cdf_diff(*self.cdf_args(beta, eta))
        # --------------------------------------------------------------
        # ll_y = sum(np.einsum('lk, ...lk -> ...',
        #                      y_i, lp_i)
        return sum(np.sum(y_i * rating_lp(xi[..., beta_ind], xi[..., eta_ind]),
                          axis=(-2, -1))
                   for (y_i, (beta_ind, eta_ind)) in zip(rating_count,
                                                         self.attribute_slices)
                   )

    def d_rating_logprob(self, xi, rating_count):
        """Gradients of rating_logprob(...),
        for each separate group of parameters,
        to be concatenated with d_situation_logprob(...)
        :param xi: 1D or 2D array of tentative parameter samples
        :param rating_count: list of 2D arrays with response counts
            rating_count[i][l, k] = number of responses for i-th ATTRIBUTE,
            at l-th ordinal level, given the k-th <=> (k0, k1, k2, ...)-th situation
        :return: d_ll = list of 1D or 2D arrays d_ll_beta, and d_ll_eta, with elements
            d_ll_beta[a][..., j] = d ln P{rating_count[a] | xi[..., :]} / d beta_a[..., j]
                w.r.t. effect parameters beta_a for a-th attribute
            d_ll_eta[t][..., j] = d ln P{rating_count[a] | xi[..., :]} / d eta_t[..., j]
                w.r.t. threshold parameters eta for t-th attribute scale
                (each set of scale thresholds may be tied to one or several attributes)
            d_ll elements ordered as groups of parameters in parameter array xi
        """
        d_ll_beta = []
        d_ll_eta = []
        for (y_i, (beta_ind, eta_ind)) in zip(rating_count,
                                              self.attribute_slices):
            beta = xi[..., beta_ind]
            eta = xi[..., eta_ind]
            (dlp_low, dlp_high) = self.rv_class.d_log_cdf_diff(*self.cdf_args(beta, eta))
            d_ll_beta.append(np.einsum('...lk, lk, jk -> ...j',
                                       dlp_low + dlp_high, y_i, - self.theta_map))
            # tensordot 2 steps?

            d_tau_d_eta = self.thr.d_tau(eta)
            d_ll_eta.append(np.einsum('...lk, lk, ...lj -> ...j',  # *** use dot , matmul ?
                                      dlp_low, y_i, d_tau_d_eta[..., :-1, :])
                            +
                            np.einsum('...lk, lk, ...lj -> ...j',
                                      dlp_high, y_i, d_tau_d_eta[..., 1:, :]))
        # all d_ll_beta, d_ll_eta elements have equal shape[:-1]
        # i.e., can be concatenated along axis=-1

        # if some eta scales tied to more than one attribute, sum those d_ll_eta elements:
        if self.tied_scales:
            d_ll_eta = [sum(d_ll_eta[i]
                            for i in tied_attribute_ind)
                        for tied_attribute_ind in self.tied_scales]
        # join into one list, elements to be concatenated by caller:
        return d_ll_beta + d_ll_eta

    def cdf_args(self, beta, eta):
        """Extract arguments for logprob calculation for ONE attribute
        :param beta: 1D or 2D array with beta parameter sample for given attribute
            beta[..., k] = ...-th sample of k-th regression-effect parameter
        :param eta: 1D or 2D array with threshold parameter sample for given attribute
            eta[..., l] = ...-th sample of l-th parameter
        :return: tuple (arg_low, arg_high) with 2D or 3D array args for probabilistic model,
            such that
            P[ l-th response | ...-th parameter sample, k-th situation ] =
            = rv_class.cdf(arg_high[..., l, k]) - rv_class.cdf(arg_low[..., l, k])
        """
        theta = np.dot(beta, self.theta_map)
        # theta[..., k] = ...-th sample of latent-variable location, given k-th situation
        # tau = response_thresholds(eta)  # INCL extreme -Inf, +Inf limits
        tau = self.thr.tau(eta)  # INCL extreme -Inf, +Inf limits
        a = tau[..., :-1, None] - theta[..., None, :]  # lower interval limits
        b = tau[..., 1:, None] - theta[..., None, :]  # upper interval limits
        return a, b

    def restrict(self, xi):
        """Adjust mean log situation-prob -> 0; -> no change of situation-prob.
        :param xi: ref to 2D array of sample parameter vectors
        :return: None
        Result: xi modified in place
            tau and associated theta samples modified by same amount
        """
        for sit_slice in self.alpha_slices:
            d = logsumexp(xi[:, sit_slice], axis=-1, keepdims=True)
            # logger.debug(f'xi situation restrict d: mean={np.mean(d):.3}; std={np.std(d):.3}; '
            #              + f'max={np.amax(d):.3}; min={np.amin(d):.3}')
            xi[:, sit_slice] -= d
        # version <= 0.9.4:
        # for (beta_slice, eta_slice) in self.attribute_slices:
        #     # beta = xi[..., beta_slice]
        #     # eta = xi[..., eta_slice]
        #     theta = np.dot(xi[..., beta_slice], self.theta_map)
        #     # tau = response_thresholds(xi[..., eta_slice])[..., 1:-1]
        #     tau = self.thr.tau(xi[..., eta_slice])[..., 1:-1]
        #     if self.restrict_attribute:
        #         d = np.mean(theta, axis=-1, keepdims=True)
        #     elif self.restrict_threshold:
        #         d = np.median(tau, axis=-1, keepdims=True)
        #     else:
        #         d = 0.
        #     # d = random offset to be zero-ed out with no effect on log-likelihood
        #     n_beta_0 = self.beta_0_size()  # **** only first regression-effect ! *****
        #     # = number of mean-location-controlling beta parameters = first part of beta_slice
        #     b0_slice = slice(beta_slice.start,
        #                      beta_slice.start + n_beta_0)
        #     # = beta slice only for FIRST regression effect term
        #     xi[:, b0_slice] -= d
        #     # eta = tau_inv(tau - d)  # **********************
        #     eta = self.thr.tau_inv(tau - d)
        #     xi[:, eta_slice] = eta
        # if self.restrict_attribute:
        #     self._restrict_attribute(xi)

    # def _restrict_attribute(self, xi):
    #     """Force mean(theta) -> 0, and adjust thresholds correspondingly
    #     :param xi: ref to 2D array of sample parameter vectors
    #     :return: None
    #     Result: xi modified in place
    #         tau and associated theta samples modified by same amount
    #     NOTE: theta and tau adjusted equally, for each attribute,
    #     so the threshold scales MUST be UNIQUE for each attribute,
    #     i.e., self.tied_scale == False
    #     """
    #     if self.tied_scales:
    #         logger.error('Cannot restrict attribute means with scale(s) tied to more than one attribute')
    #         # *** should not happen, checked earlier
    #     for (beta_slice, eta_slice) in self.attribute_slices:
    #         # beta = xi[..., beta_slice]
    #         # eta = xi[..., eta_slice]
    #         theta = np.dot(xi[..., beta_slice], self.theta_map)
    #         tau = self.thr.tau(xi[..., eta_slice])[..., 1:-1]
    #         d = np.mean(theta, axis=-1, keepdims=True)
    #         # d = random offset to be zero-ed out with no effect on log-likelihood
    #         n_beta_0 = self.beta_0_size()  # **** only first regression-effect ! *****
    #         # = number of mean-location-controlling beta parameters = first part of beta_slice
    #         b0_slice = slice(beta_slice.start,
    #                          beta_slice.start + n_beta_0)
    #         # = beta slice only for FIRST regression effect term
    #         xi[:, b0_slice] -= d
    #         # eta = tau_inv(tau - d)  # **********************
    #         eta = self.thr.tau_inv(tau - d)
    #         xi[:, eta_slice] = eta

    # ------------------------------------ Help methods for ema_respondent.initialize():
    def initialize_xi(self, z, y):
        """Initialize individual parameter vector, given observed counts
        :param z: 2D array with
            z[k0, k] = number of EMA records at k0-th test phase
            in k-th <=> (k1, k2, ...)-th situation, EXCL. k0= test phase
        :param y: list of 2D arrays,
            y[i][l, k] = number of responses at l-th ordinal level for i-th ATTRIBUTE question
            given k-th <=> (k0, k1, k2, ...)-th situation (INCL. k0 = test phase)
        :return: xi = 1D array with all parameters,
            ordered by parameter sub-types (alpha, beta, eta) as defined by self
        """
        alpha = list(_initialize_situation_param(z))
        eta = [_initialize_rating_eta(y_i, self.thr)
               for y_i in y]
        # beta = [_initialize_rating_beta(y_i, eta_i, self.theta_map)
        #         for (y_i, eta_i) in zip(y, eta)]
        beta = [_initialize_rating_beta(y_i, self.thr.tau(eta_i), self.theta_map)
                for (y_i, eta_i) in zip(y, eta)]
        if self.tied_scales:  # ****************** sum y instead ! ****************
            eta = [sum(eta[i] for i in tied_attribute_ind)
                   for tied_attribute_ind in self.tied_scales]
        for eta_scale in eta:
            eta_scale -= np.sum(eta_scale)
        # just in case _initial_rating_eta yields non-zero values
        # order as alpha[0], alpha[1],..., beta[0], beta[1], ..., eta[0], eta[1],...
        return np.concatenate([*chain(alpha, beta, eta)])


# ---------------------------------------- Other support functions:

def _initialize_situation_param(z):
    """Crude initial estimate of situation logprob parameters
    :param z: 2D array with situation counts
        z[t, k] = n EMA records in k-th <-> (k1, k2,...)-th situation
            in t-th test phase
    :return: alpha = 2D array
        alpha[t, k] = log prob of k-th situation at t-th test phase
        alpha.shape == (z.shape[0], z[0].size)
    """
    p = z + PRIOR_PSEUDO_RESPONDENT
    p /= np.sum(p, axis=-1, keepdims=True)
    return np.log(p)


def _initialize_rating_eta(y, thr):
    """Crude initial estimate of threshold-defining parameters for ONE attribute question
    :param y: 2D rating_count array,
        y[l, k] = number of responses at l-th ordinal level,
        given the k-th <-> (k0, k1, ...)-th situation category
    :param thr: Thresholds subclass, to determine number of eta parameters
    :return: eta = 1D array of eta parameters, defining thresholds as
        tau = ema_base.response_thresholds(eta)
        eta.shape == y.shape[:1]
    Method: set eta parameters to match observed relative freq of all rating grades
        as if theta = 0.
    OR just set all eta == 0.
    """
    # p = np.sum(y, axis=-1) + JEFFREYS_CONC
    # p /= np.sum(p)
    # return mapped_width_inv(p)
    # *** This method seems to improve the first LL,
    # but sometimes give worse or same long-term result
    return np.zeros(thr.n_param(len(y)))  # original simple method


def _initialize_rating_beta(y, tau, theta_map):
    """Crude initial estimate of regression-effect parameters for ONE attribute question
    :param y: 2D rating_count array,
        y[l, k] = number of responses at l-th ordinal level,
        given the k-th <-> (k0, k1, ...)-th situation category
    :param tau: 1D array with threshold parameters, incl extreme -inf + inf
    :return: beta = array of sensory-variable locations theta, such that
        theta approx = np.dot(beta, theta_map),
        theta[k] = estimated location of latent variable,
        given the k-th <-> (k0, k1, ...)-th situation category
    """
    expit_tau = expit(tau)
    typical_theta = logit((expit_tau[:-1] + expit_tau[1:]) / 2)
    # = back-transformed midpoints in each response interval
    # typical_theta[l] = theta given l-th response
    p = y + PRIOR_PSEUDO_RESPONDENT
    p /= np.sum(p, axis=0, keepdims=True)
    theta = np.dot(typical_theta, p)
    # = typical (midpoint) locations, given y
    beta = np.linalg.lstsq(theta_map.T, theta, rcond=None)[0]
    # -> theta approx = np.dot(beta, theta_map),
    return beta


def _make_prior(n_parameters,
                alpha_slices,
                attribute_slices,
                rv_class):
    """Create hyperprior for parameter distribution in the total population
    :param n_parameters: total number of parameters
    :param alpha_slices: list with index slice for each situation phase
    :param attribute_slices: list with one AttributeSlice tuple for each attribute
    :return: single GaussianRV instance
    """
    prec_a = PRIOR_PREC_A  # * np.ones(n_parameters)  # still scalar, same for all elements
    prec_b = PRIOR_PREC_B * np.ones(n_parameters)
    for (b_slice, e_slice) in attribute_slices:
        prec_b[b_slice] *= rv_class.scale**2
        # prec_a[e_slice] *= 100.  # ****** temp fix just for TEST
        # ************* TEST reduce prior threshold-parameter variance, no
    loc = np.zeros(n_parameters)
    for a_slice in alpha_slices:
        loc[a_slice] = - np.log(a_slice.stop - a_slice.start)
    return GaussianRV.initialize(loc=loc,
                                 learned_weight=PRIOR_PSEUDO_RESPONDENT,
                                 prec_a=prec_a,
                                 prec_b=prec_b)


def _check_effects(emf, effects):
    """Check that all effects refer to unique situation keys
    :param emf: ema_data.EmaFrame instance
    :param effects: iterable with regression-effect specifications
        effects[i] = single key in emf.situation_dtypes.keys(), OR tuple of such keys
    :return: None
    Result: raise RuntimeError if incorrect effects
    """
    effect_keys = sum((list(e_i) if type(e_i) is tuple else [e_i]
                      for e_i in effects), start=[])
    # = flat list with all situation keys from effects
    if len(set(effect_keys)) != len(effect_keys):
        raise RuntimeError('situation keys can occur only ONCE in regression effects')
    for sit_key in emf.situation_dtypes.keys():
        if len(emf.situation_dtypes[sit_key].categories) > 1 and sit_key not in effect_keys:
            logger.warning(f'situation dim {repr(sit_key)} to be disregarded in regression?')
    for e_i in effect_keys:
        if e_i not in emf.situation_dtypes.keys():
            raise RuntimeError(f'regression effect key {e_i} is not a situation key')


def _theta_map(emf, effects, restrict_attr):
    """Create 2D array for extraction of attribute locations
    from parameter vector samples
    :param emf: ema_data.EmaFrame instance,
        defining the analysis model structure.
    :param effects: iterable with regression-effect specifications
        effects[i] = tuple with one or more key(s) from emf.situation_dtypes.keys()
    :param restrict_attr: boolean switch
        to force mean theta across situations -> 0., regardless of beta
    :return: th = 2D array, such that
        theta[..., :] = np.dot(beta[..., :], th) is ...-th sample of attribute location
            in the :-th <=> (k0, k1, ...)-th situation, with
        beta[..., j] = xi[:, attr_slice][..., j] is the j-th regression effect parameter,
        and xi is the array of row parameter-vector samples.
    """
    def theta_one(effect):
        """Make theta_map part for ONE effect term
        :param effect: tuple of one or more situation keys
        :return: mD array th, with
            th[j, k0, k1, ...] = j-th effect given (k0, k1, ...)-th situation
            th[j, 0, 0, ...] = 0, in 0-th reference situation.
            th.shape == (size_effect, *emf.situation_shape), where
            size_effect = size of category array defined by effect.
        """
        beta_shape = tuple(len(emf.situation_dtypes[sc_i].categories) for sc_i in effect)
        # = situation_shape, for this effect term
        beta_ndim = len(beta_shape)
        beta_size = int(np.prod(beta_shape))
        t = np.eye(beta_size).reshape((beta_size, *beta_shape))
        if restrict_attr or effect != effects[0]:
            t = t[1:]
            # if restrict_attr: only relative effects re. first category, delete first
            # else: include absolute theta intercept parameter only in FiRST effect term,
            # with same influence on theta in all effect terms
        # expand t.ndim to number of situation dimensions:
        t = t.reshape(t.shape + tuple(np.ones(len(emf.situation_dtypes) - beta_ndim,
                                              dtype=int)
                                      ))
        ax_0 = range(1, 1 + len(beta_shape))
        ax_new = tuple(list(emf.situation_dtypes).index(e_i) + 1
                       for e_i in effect)
        t = np.moveaxis(t, ax_0, ax_new)
        return t + np.zeros(emf.situation_shape)
    # ---------------------------------------------------------------
    th = np.concatenate([theta_one(e_tuple)
                         for e_tuple in effects],
                        axis=0)
    th = th.reshape((th.shape[0], -1))  # keep as 2D for simplicity
    if restrict_attr:
        th -= np.mean(th, axis=-1, keepdims=True)
        # forcing mean(theta) -> 0., regardless of regression beta parameters
    return th


# ------------------------------------------------- TEST:
if __name__ == '__main__':
    # from scipy.optimize import approx_fprime, check_grad

    from ema_data import EmaFrame
    from ema_latent import Bradley

    print('*** Testing some ema_base module functions ')
    emf = EmaFrame.setup(situations={'CoSS': [f'C{i}' for i in range(1, 8)],
                                     'Viktigt': ('Lite',
                                                 'Ganska',
                                                 'Mycket'),
                                     # 'Test': (None, None),
                                     'HA': ('A', 'B')
                                     }, attributes={'Speech': ('Difficult',
                                                               'Normal',
                                                               'Easy'),
                                                    'Quality': ('Bad', 'Good')
                                                    })

    # print('emf=\n', emf)

    # Testing restrictions (both cannot be true, but both may be False):
    restrict_threshold = False   # False
    restrict_attribute = True   # False
    print(f'restrict_attribute= {restrict_attribute}; restrict_threshold= {restrict_threshold}')
    main_theta = {'Phase': [0.],
                  'CoSS': 0.1 * np.arange(len(emf.situation_dtypes['CoSS'].categories)),
                  'Viktigt': np.array([-1., 0., 1.]),
                  'HA': np.array([0., 1.])
                  }
    # = only main effects, additive
    regr_effects = ['HA', ('CoSS', 'Viktigt')]
    # regr_effects = [('CoSS', 'Viktigt'), 'HA']

    # effect_HA = np.array([-1., 1.])
    # effect_CoSS = 0.1 * np.arange(len(emf.situation_dtypes['CoSS']))
    # effect_Vikt = np.array([-1., 0., 1.])
    true_theta = main_theta['CoSS'][:, None] + main_theta['Viktigt']
    true_theta = true_theta[..., None] + main_theta['HA']
    true_theta = true_theta[None, ...]  # only one phase
    # stored like emf.situation_dtypes
    # true_theta[t, c, v, h] = theta in t-th phase, c-th CoSS, v-th Vikt, h-th HA

    theta_HA = main_theta['HA']
    beta_HA = theta_HA[1:] - theta_HA[0]  # rel. first HA
    theta_CoSS_Vikt = main_theta['CoSS'][:, None] + main_theta['Viktigt']
    theta_CoSS_Vikt = theta_CoSS_Vikt.reshape((-1,))
    beta_CoSS_Vikt = theta_CoSS_Vikt[1:] - theta_CoSS_Vikt[0]  # rel. first CoSS_Vikt case
    if restrict_attribute:
        # only relative beta differences:
        beta_all = np.concatenate((beta_HA, beta_CoSS_Vikt))
        true_theta -= np.mean(true_theta) # force mean theta -> 0.
    else:
        # include absolute theta for HA:
        beta_all = np.concatenate((theta_HA + theta_CoSS_Vikt[0], beta_CoSS_Vikt))
    # for regr_effects = ['HA', ('CoSS', 'Viktigt')]
    # beta_all = np.concatenate((theta_CoSS_Vikt.reshape((-1)), beta_HA))
    # for regr_effects = [('CoSS', 'Viktigt'), 'HA']

    if restrict_threshold:
        thr = ThresholdsMidFixed
    else:
        thr = ThresholdsFree
    eta = [np.zeros(thr.n_param(len(r_cat.categories)))
           for r_cat in emf.attribute_dtypes.values()]

    alpha = np.zeros(emf.situation_shape).reshape((-1))
    print(f'len(alpha)= {len(alpha)}')
    print('; '.join(f'{a}: len(beta)= {len(beta_all)}'
                    for a in emf.attribute_dtypes.keys())
          )
    print('; '.join(f'{a}: len(eta)= {len(eta_a)}'
                    for (a, eta_a) in zip(emf.attribute_dtypes.keys(),
                                          eta)
                    )
          )

    # xi = np.concatenate((lp_situations, *beta_eta))
    xi = np.concatenate((alpha,
                         np.tile(beta_all, len(emf.attribute_dtypes)),
                         *eta))
    print(f'n_parameters= {len(xi)}')
    xi = xi[None, :]
    print(f'xi= {xi}')

    p_base = EmaParamBase.initialize(emf, regr_effects, rv_class=Bradley,
                                     restrict_threshold=restrict_threshold,
                                     restrict_attribute=restrict_attribute)
    print('p_base= ', p_base)
    print(f'p_base.n_parameters = {p_base.n_parameters}')
    print(f'p_base.theta_map.shape= {p_base.theta_map.shape}')

    print('\n*** Testing param extraction methods ***')

    for (phase, alpha_slice) in zip(emf.situation_dtypes[emf.phase_key].categories,
                                    p_base.alpha_slices):
        print(f'{emf.phase_key}= {phase}: xi[:, alpha_slice].shape= ', xi[:, alpha_slice].shape)
    # for (phase, alpha) in zip(emf.situation_dtypes[emf.phase_key], p_base.situation_logprob(xi)):
    #     print(f'{emf.phase_key}= {phase}: prob-mass=\n', np.exp(alpha))
    print(f'\nemf.situation_shape= {emf.situation_shape}')
    # for (a, beta_slice) in zip(emf.attribute_grades.keys(),
    #                            p_base.beta_slices):
    # for (a, a_slice) in p_base.attribute_slices.items():
    for (a, (beta_slice, eta_slice)) in zip(emf.attribute_dtypes.keys(),
                                            p_base.attribute_slices):
        print(f'\nAttribute {a}.beta_slice: {beta_slice}')
        print(f'\nAttribute {a}: xi[:, beta_slice]={xi[:, beta_slice]}')

    for a in emf.attribute_dtypes.keys():
        theta = p_base.attribute_theta(xi, a)
        print(f'\nAttribute {a}: theta.shape={theta.shape}.')
        sc_dim_0 = 'Phase'
        sc_dim_1 = 'CoSS'
        theta = theta.reshape((-1, *emf.situation_shape))  # *** just for testing
        for (sc_0, theta_0, true_theta_0) in zip(emf.situation_dtypes[sc_dim_0].categories,
                                                 theta[0], true_theta):
            print(f'\t{sc_dim_0}= {sc_0}:')
            for (sc_1, theta_1, true_theta_1) in zip(emf.situation_dtypes[sc_dim_1].categories,
                                                     theta_0, true_theta_0):
                print(f'\t\t{sc_dim_1}= {sc_1}: theta=\n', theta_1)
                print(f'\t\t{sc_dim_1}= {sc_1}: true_theta=\n', true_theta_1)

    for a in emf.attribute_dtypes.keys():
        tau = p_base.attribute_tau(xi, a)
        print(f'\nAttribute {a}: tau.shape={tau.shape}. tau=\n', tau)

    xi[0, 0:2] = 2.
    print('\n***Testing situation_prob')
    p_ds = p_base.situation_prob_df(xi, groupby=('Viktigt',))
    print('situation prob.=\n',p_ds)
    print('\n***Testing attribute_theta:')
    th_ds = p_base.attribute_theta_df(xi, a='Speech')
    print(th_ds)
    p_ds = p_base.situation_prob_df(xi)
    print('situation prob.=\n',p_ds)
    th_ds = p_base.attribute_theta_df(xi, a='Speech', groupby=('CoSS',))
    print('Speech=\n', th_ds)


